import sys

import chainer
import numpy as np

from onnx_chainer.functions.array import get_slice_node
from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


@support((1, 6, 7))
def convert_BatchNormalization(
        func, opset_version, input_names, output_names, context):
    is_fixed_bn = len(func.inputs) > 3

    # NOTE: even if `use_beta=False` or `use_gamma=False`, beta or gamma
    # are set in inputs by RetainHook,
    beta_param = func.inputs[2].get_variable_or_none()
    gamma_param = func.inputs[1].get_variable_or_none()
    namedlink = context.get_link(beta_param) or context.get_link(gamma_param)

    if namedlink is not None:
        prefix, link = namedlink
        if is_fixed_bn:
            mean = link.avg_mean
            var = link.avg_var
        else:
            # on train mode, avg_mean would be updated, so make them from x
            x = func.inputs[0].get_variable().array
            mean = x.mean(axis=func.axis)
            var = x.var(axis=func.axis)
    else:
        prefix = None
        if is_fixed_bn:
            mean = func.inputs[3].get_variable().array
            var = func.inputs[4].get_variable().array
        else:
            x = func.inputs[0].get_variable().array
            mean = x.mean(axis=func.axis)
            var = x.var(axis=func.axis)

    def add_param(v, suffix):
        if prefix is None:
            return context.add_param(v, suffix)
        else:
            return context.add_param(
                v, '{}_{}'.format(prefix, suffix), use_original_name=True)

    if is_fixed_bn:
        if context.implicit_inputs.pop(input_names[3], None) is not None:
            mean_name = add_param(mean, 'avg_mean')
            input_names[3] = mean_name
        if context.implicit_inputs.pop(input_names[4], None) is not None:
            var_name = add_param(var, 'avg_var')
            input_names[4] = var_name
    else:
        maen_name = add_param(mean, 'avg_mean')
        var_name = add_param(var, 'avg_var')
        input_names.extend([maen_name, var_name])

    momentum = getattr(func, 'decay', 0.)

    # TODO(disktnk): On definition of ONNX's BatchNormalization operator,
    # outputs one required output and four optional outputs. This converter
    # must make 5 values for output and return them.

    if opset_version == 1:
        return onnx_helper.make_node(
            'BatchNormalization', input_names, output_names,
            epsilon=func.eps,
            momentum=momentum,
            is_test=not chainer.config.train,
            consumed_inputs=[False, False, False, True, True],
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'BatchNormalization', input_names, output_names,
            epsilon=func.eps,
            momentum=momentum,
            is_test=not chainer.config.train,
        ),
    elif opset_version == 7:
        return onnx_helper.make_node(
            'BatchNormalization', input_names, output_names,
            epsilon=func.eps,
            momentum=momentum,
        ),


@support((1, 6, 7))
def convert_FixedBatchNormalization(
        func, opset_version, input_names, output_names, context):
    return convert_BatchNormalization(
        func, opset_version, input_names, output_names, context)


@support((5, 10))
def convert_GroupNormalization(
        func, opset_version, input_names, output_names, context):
    # drop opset < 5, to reduce supporting cost of old Reshape op
    # calculation process is from
    # https://github.com/chainer/chainer/blob/v6.2.0/chainer/functions/normalization/group_normalization.py  # NOQA
    # support dynamic batch size, but channel size is expected to be fixed

    group = context.add_const(np.array(func.groups, dtype=np.int64), 'group')
    eps = context.add_const(np.array(func.eps, dtype=np.float32), 'eps')
    channel_size = context.add_const(
        np.array([func.inputs[0].shape[1]], dtype=np.int64), 'channel')

    neg_one = context.add_const(np.array([-1], dtype=np.int64), 'neg_one')

    gb = onnx_helper.GraphBuilder()

    # make reduced input
    original_shape = gb.op('Shape', [input_names[0]])
    batch_size = get_slice_node(
        gb, opset_version, context, [original_shape], [0], [0], [1])
    batched_group = gb.op('Mul', [batch_size, group])
    reduce_shape = gb.op('Concat', [batched_group, neg_one], axis=0)
    reduced_x = gb.op('Reshape', [input_names[0], reduce_shape])

    # calculate mean, var and x_hat
    mean = gb.op('Unsqueeze', [
        gb.op('ReduceMean', [reduced_x], axes=[1], keepdims=0)], axes=[1])
    x_hat = gb.op('Sub', [reduced_x, mean])
    var = gb.op('Add', [
        gb.op(
            'ReduceMean', [gb.op('Mul', [x_hat, x_hat])], axes=[1],
            keepdims=0),
        eps])
    inv_std = gb.op('Unsqueeze', [
        gb.op('Reciprocal', [gb.op('Sqrt', [var])])], axes=[1])
    x_hat_ = gb.op('Mul', [x_hat, inv_std])

    # make out y
    groupless_shape = gb.op(
        'Concat', [batch_size, channel_size, neg_one], axis=0)
    y_org = gb.op('Reshape', [x_hat_, groupless_shape])

    # gamma/beta
    gamma = gb.op('Unsqueeze', [input_names[1]], axes=[1])
    beta = gb.op('Unsqueeze', [input_names[2]], axes=[1])
    y_g = gb.op('Mul', [y_org, gamma])
    y_b = gb.op('Add', [y_g, beta])
    gb.op('Reshape', [y_b, original_shape])

    return gb.nodes(output_names)


def convert_LocalResponseNormalization(
        func, opset_version, input_names, output_names, context):
    size = int(func.n)
    return onnx_helper.make_node(
        'LRN', input_names, output_names,
        alpha=float(func.alpha) * size,
        beta=float(func.beta),
        bias=float(func.k),
        size=size,
    ),


def convert_NormalizeL2(
        func, opset_version, input_names, output_names, context):
    if isinstance(func.axis, tuple) and len(func.axis) != 1:
        raise ValueError(
            'Normalization along with multiple axes ({}) are not supported in '
            'the ONNX\'s LpNormalization operator.'.format(func.axis))
    if abs(func.eps - 1e-5) > sys.float_info.epsilon:
        # default value of F.normaize eps is 1e-5
        raise ValueError(
            '\'eps\' is not supported in the ONNX\'s LpNormalization operator,'
            ' so that ONNX-Chainer does not accept custom values for \'eps\' '
            '({})'.format(func.eps))

    return onnx_helper.make_node(
        'LpNormalization', input_names, output_names,
        axis=int(func.axis[0]),
        p=2,
    ),



import chainer
from chainer.backends import cuda
from chainer import distribution
from chainer.functions.math import exponential
import chainer.functions.math.sum as sum_mod
from chainer.utils import cache


def _stack(xp, xs, axis):
    try:
        return xp.stack(xs, axis)
    except AttributeError:
        # in case numpy<1.10, which does not have numpy.stack
        return xp.concatenate(
            [xp.expand_dims(x, axis) for x in xs],
            axis=axis)


def _random_choice(xp, a, size, p):
    try:
        return xp.random.choice(a, size, p=p)
    except ValueError:
        # Validate the sum of the probabilities as NumPy PR #6131 (numpy>=1.10)
        tol = xp.finfo(p.dtype).eps ** 0.5
        p = p.astype(xp.float64)
        xp.testing.assert_allclose(p.sum(), 1, rtol=0, atol=tol)

        # Normalize the probabilities as they satisfy the validation above, and
        # generate samples again
        p /= p.sum()
        return xp.random.choice(a, size, p=p)


class OneHotCategorical(distribution.Distribution):

    """OneHotCategorical Distribution.

    Args:
        p(:class:`~chainer.Variable` or :ref:`ndarray`): Parameter of
            distribution.
    """

    def __init__(self, p):
        super(OneHotCategorical, self).__init__()
        self.__p = p

    @cache.cached_property
    def p(self):
        return chainer.as_variable(self.__p)

    @cache.cached_property
    def log_p(self):
        return exponential.log(self.p)

    @property
    def batch_shape(self):
        return self.p.shape[:-1]

    @property
    def event_shape(self):
        return self.p.shape[-1:]

    @property
    def _is_gpu(self):
        return isinstance(self.p.data, cuda.ndarray)

    def log_prob(self, x):
        return sum_mod.sum(self.log_p * x, axis=-1)

    @cache.cached_property
    def mean(self):
        return self.p

    @property
    def params(self):
        return {'p': self.p}

    def sample_n(self, n):
        xp = chainer.backend.get_array_module(self.p)
        obo_p = self.p.data.reshape((-1,) + self.event_shape)
        eye = xp.eye(self.event_shape[0], dtype=self.p.dtype)
        eps = [_random_choice(xp, one_p.shape[0], size=(n,), p=one_p)
               for one_p in obo_p]
        eps = _stack(xp, eps, axis=1).reshape((n,)+self.batch_shape)
        eps = eye[eps]
        noise = chainer.Variable(eps)
        return noise

    @cache.cached_property
    def variance(self):
        return self.p * (1. - self.p)


@distribution.register_kl(OneHotCategorical, OneHotCategorical)
def _kl_one_hot_categorical_one_hot_categorical(dist1, dist2):
    return sum_mod.sum(dist1.p * (dist1.log_p - dist2.log_p), axis=-1)



import chainer


def get_device(device_id=None, use_chainerx=False):
    """Get device object

    Currently in Chainer, there are 3 officially-supported backends
    (numpy, cupy, and chainerx) and 2 devices (CPU and NVIDIA GPUs).
    Also, ChainerX has its own backend system, so there are 4 combinations
    (numpy, cupy, chainerx+native, chainerx+cuda). This utility function
    is a boilerplate to get device object in ChainerMN contexts.
    """
    if device_id is not None:
        if use_chainerx:
            device = 'cuda:{}'.format(device_id)
        else:
            device = '@cupy:{}'.format(device_id)
    else:
        if use_chainerx:
            device = 'native:0'
        else:
            device = '@numpy'

    return chainer.get_device(device)



import numpy as np

from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


def convert_Convolution2DFunction(
        func, opset_version, input_names, output_names, context):
    if hasattr(func, 'dy') and hasattr(func, 'dx'):
        node = onnx_helper.make_node(
            'Conv', input_names, output_names,
            dilations=(func.dy, func.dx),
            kernel_shape=func.inputs[1].shape[2:],
            # pads: [x1_begin, x2_begin...x1_end, x2_end,...]
            pads=(func.ph, func.pw, func.ph, func.pw),
            strides=(func.sy, func.sx),
            group=func.groups,
        )
    else:
        node = onnx_helper.make_node(
            'Conv', input_names, output_names,
            dilations=(1, 1),
            kernel_shape=func.inputs[1].shape[2:],
            pads=(func.ph, func.pw, func.ph, func.pw),
            strides=(func.sy, func.sx),
            group=func.groups,
        )
    return node,


def convert_ConvolutionND(
        func, opset_version, input_names, output_names, context):
    pad = []
    x_ndim = len(func.inputs[0].shape)
    w_ndim = len(func.inputs[1].shape)
    for _ in range(x_ndim - w_ndim):
        pad.append(0)
    for p in func.pad:
        pad.append(p)
    pad = pad * 2

    return onnx_helper.make_node(
        'Conv', input_names, output_names,
        kernel_shape=func.inputs[1].shape[2:],
        pads=pad,
        strides=func.stride,
        group=func.groups,
    ),


def convert_Deconvolution2DFunction(
        func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node(
        'ConvTranspose', input_names, output_names,
        kernel_shape=func.inputs[1].shape[2:],
        output_shape=(func.outh, func.outw),
        # pads: [x1_begin, x2_begin...x1_end, x2_end,...]
        pads=(func.ph, func.pw, func.ph, func.pw),
        strides=(func.sy, func.sx),
        group=func.groups,
    ),


def convert_DeconvolutionND(
        func, opset_version, input_names, output_names, context):
    pad = []
    x_ndim = len(func.inputs[0].shape)
    w_ndim = len(func.inputs[1].shape)
    for _ in range(x_ndim - w_ndim):
        pad.append(0)
    for p in func.pad:
        pad.append(p)
    pad = pad * 2

    return onnx_helper.make_node(
        'ConvTranspose', input_names, output_names,
        kernel_shape=func.inputs[1].shape[2:],
        output_shape=func.outs,
        pads=pad,
        strides=func.stride,
        group=func.groups,
    ),


def convert_EmbedIDFunction(
        func, opset_version, input_names, output_names, context):
    x_index_name, W_name = input_names
    input_names = [W_name, x_index_name]

    if func.ignore_label is not None:
        raise ValueError(
            'Current ONNX doesn\'t support ignore_label for EmbedID.')

    return onnx_helper.make_node(
        'Gather', input_names, output_names, axis=0),


@support((1, 6, 7))
def convert_LinearFunction(
        func, opset_version, input_names, output_names, context):
    # When the func has bias
    if len(func.inputs) == 2:
        bias_dim = func.inputs[1].shape[0]
        bias = np.zeros((bias_dim,), dtype=func.inputs[0].dtype)
        bias_name = context.add_param(bias, 'bias')
        input_names.append(bias_name)

    if opset_version == 1 or opset_version == 6:
        return onnx_helper.make_node(
            'Gemm', input_names, output_names,
            alpha=1.0, beta=1.0, broadcast=1, transA=0, transB=1),
    elif opset_version == 7:
        return onnx_helper.make_node(
            'Gemm', input_names, output_names,
            alpha=1.0, beta=1.0, transA=0, transB=1),



import copy

import chainer
import chainermn


def create_mnbn_model(link, comm, communication_backend='auto'):
    """Create a link object with MultiNodeBatchNormalization.

    Returns a copy of `link`, where BatchNormalization is replaced
    by MultiNodeBatchNormalization.

    Args:
        link: Link object
        comm: ChainerMN communicator
        communication_backend (str): ``mpi``, ``nccl`` or ``auto``. It is used
            to determine communication backend of MultiNodeBatchNormalization.
            If ``auto``, use the best communication backend for each
            communicator.

    Returns:
        Link object where BatchNormalization is replaced
        by MultiNodeBatchNormalization.

    """

    if isinstance(link, chainer.links.BatchNormalization):
        mnbn = chainermn.links.MultiNodeBatchNormalization(
            size=link.avg_mean.shape,
            comm=comm,
            decay=link.decay,
            eps=link.eps,
            dtype=link.avg_mean.dtype,
            use_gamma=hasattr(link, 'gamma'),
            use_beta=hasattr(link, 'beta'),
            communication_backend=communication_backend,
        )
        mnbn.copyparams(link)
        for name in link._persistent:
            mnbn.__dict__[name] = copy.deepcopy(link.__dict__[name])
        return mnbn
    elif isinstance(link, chainer.Chain):
        new_children = [
            (child_name, create_mnbn_model(link.__dict__[child_name], comm,
                                           communication_backend))
            for child_name in link._children
        ]
        new_link = copy.deepcopy(link)
        for name, new_child in new_children:
            new_link.__dict__[name] = new_child
        return new_link
    elif isinstance(link, chainer.Sequential):
        new_link = copy.deepcopy(link)
        for i, l in enumerate(link):
            new_l = create_mnbn_model(l, comm, communication_backend)
            new_link[i] = new_l
        return new_link
    elif isinstance(link, chainer.ChainList):
        new_children = [
            create_mnbn_model(l, comm, communication_backend) for l in link]
        new_link = copy.deepcopy(link)
        for i, new_child in enumerate(new_children):
            new_link._children[i] = new_child
        return new_link
    else:
        return copy.deepcopy(link)



import functools

import chainer
from chainer import backend
from chainer.testing import _bundle
from chainer.testing import attr
import chainerx


# TODO(hvy): BackendConfig.__enter__ does not have to modify the current
# device. Change it so that it does not.
class BackendConfig(object):

    _props = [
        # ChainerX
        ('use_chainerx', False),
        ('chainerx_device', None),
        # CuPy
        ('use_cuda', False),
        ('cuda_device', None),  # 0 by default, if use_cuda=True
        ('use_cudnn', 'never'),
        ('cudnn_deterministic', False),
        ('autotune', False),
        ('cudnn_fast_batch_normalization', False),
        # Intel64
        ('use_ideep', 'never'),
    ]

    _device = None

    def __init__(self, params):
        if not isinstance(params, dict):
            raise TypeError('params must be a dict.')
        self._contexts = []

        # Default values
        for k, v in self._props:
            setattr(self, k, v)
        # Specified values
        for k, v in params.items():
            if not hasattr(self, k):
                raise ValueError('Parameter {} is not defined'.format(k))
            setattr(self, k, v)

        self._check_params()
        self._adjust_params()

    def _check_params(self):
        # Checks consistency of parameters

        if self.use_chainerx:
            assert isinstance(self.chainerx_device, str), (
                '\'chainerx_device\' parameter is expected to be a string '
                'representing a ChainerX device specifier')

    def _adjust_params(self):
        # Adjusts parameters, e.g. fill the default values
        if self.use_cuda:
            if self.cuda_device is None:
                self.cuda_device = 0

    @property
    def xp(self):
        return self.device.xp

    @property
    def device(self):
        if self._device is None:
            if self.use_cuda:
                device = backend.GpuDevice.from_device_id(self.cuda_device)
            elif self.use_chainerx:
                device = backend.ChainerxDevice(
                    chainerx.get_device(self.chainerx_device))
            elif self.use_ideep != 'never':
                device = backend.Intel64Device()
            else:
                device = backend.CpuDevice()
            self._device = device
        return self._device

    def __enter__(self):
        contexts = [
            chainer.using_config(
                'use_cudnn', self.use_cudnn),
            chainer.using_config(
                'cudnn_deterministic', self.cudnn_deterministic),
            chainer.using_config(
                'autotune', self.autotune),
            chainer.using_config(
                'use_ideep', self.use_ideep),
            chainer.using_device(self.device),
        ]
        for c in contexts:
            c.__enter__()
        self._contexts.append(contexts)
        return self

    def __exit__(self, typ, value, traceback):
        contexts = self._contexts.pop()
        for c in reversed(contexts):
            c.__exit__(typ, value, traceback)

    def __repr__(self):
        lst = []
        for k, _ in self._props:
            lst.append('{}={!r}'.format(k, getattr(self, k)))
        return '<BackendConfig {}>'.format(' '.join(lst))

    def get_func_str(self):
        """Returns a string that can be used in method name"""
        lst = []
        for k, _ in self._props:
            val = getattr(self, k)
            if val is True:
                val = 'true'
            elif val is False:
                val = 'false'
            else:
                val = str(val)
            lst.append('{}_{}'.format(k, val))
        return '__'.join(lst)

    def get_pytest_marks(self):
        marks = []
        if self.use_chainerx:
            marks.append(attr.chainerx)
            backend_name, device_index = self.chainerx_device.split(':')
            device_index = int(device_index)
            if backend_name == 'cuda':
                marks.append(attr.gpu)
                if device_index >= 1:
                    marks.append(attr.multi_gpu(device_index + 1))
        elif self.use_cuda:
            marks.append(attr.gpu)
            if self.use_cudnn != 'never':
                marks.append(attr.cudnn)
            if self.cuda_device >= 1:
                marks.append(attr.multi_gpu(self.cuda_device + 1))
        else:
            if self.use_ideep != 'never':
                marks.append(attr.ideep)

        assert all(callable(_) for _ in marks)
        return marks

    def get_array(self, np_array):
        return self.device.send(np_array)


def _test_case_generator(base, method_names, params):
    # Defines the logic to generate test case classes parameterized with
    # backends.

    if method_names is not None:
        def method_generator(base_method):
            if base_method.__name__ in method_names:
                return None
            return base_method

        yield (base.__name__, {}, method_generator)

    for i_param, param in enumerate(params):
        backend_config = BackendConfig(param)
        marks = backend_config.get_pytest_marks()
        cls_name = '{}_{}'.format(base.__name__, backend_config.get_func_str())

        def method_generator(base_method):
            # Generates a wrapped test method

            if (method_names is not None
                    and base_method.__name__ not in method_names):
                return None

            # Bind to a new variable.
            backend_config2 = backend_config

            @functools.wraps(base_method)
            def new_method(self, *args, **kwargs):
                return base_method(self, backend_config2, *args, **kwargs)

            # Apply test marks
            for mark in marks:
                new_method = mark(new_method)

            return new_method

        yield (cls_name, {}, method_generator)


def inject_backend_tests(method_names, params):
    if not (method_names is None or isinstance(method_names, list)):
        raise TypeError('method_names must be either None or a list.')
    if not isinstance(params, list):
        raise TypeError('params must be a list of dicts.')
    if not all(isinstance(d, dict) for d in params):
        raise TypeError('params must be a list of dicts.')

    return _bundle.make_decorator(
        lambda base: _test_case_generator(base, method_names, params))



import typing as tp

import numpy

from chainer import function_node


# TODO(okapies): Split this into independent .py and .pyi files
# mypy can't handle wildcard imports properly, so I aggregate
# the signatures under chainerx module to this place.
# See https://github.com/python/mypy/issues/5908 for the details


# chainerx_cc/chainerx/python/backend.cc
class Backend:
    @property
    def context(self) -> Context: ...

    @property
    def name(self) -> str: ...

    def get_device(self, arg0: int) -> Device: ...

    def get_device_count(self) -> int: ...


# chainerx_cc/chainerx/python/backprop_mode.cc
class NoBackpropMode:
    def __enter__(self) -> None: ...

    def __exit__(self, *args) -> None: ...


class ForceBackpropMode:
    def __enter__(self) -> None: ...

    def __exit__(self, *args) -> None: ...


@tp.overload
def no_backprop_mode(arg0: BackpropId) -> NoBackpropMode: ...


@tp.overload
def no_backprop_mode(arg0: tp.List[BackpropId]) -> NoBackpropMode: ...


@tp.overload
def no_backprop_mode(context: tp.Optional[Context]=None) -> NoBackpropMode: ...


@tp.overload
def force_backprop_mode(arg0: BackpropId) -> ForceBackpropMode: ...


@tp.overload
def force_backprop_mode(arg0: tp.List[BackpropId]) -> ForceBackpropMode: ...


@tp.overload
def force_backprop_mode(context: Context=None) -> ForceBackpropMode: ...


@tp.overload
def is_backprop_required(arg0: BackpropId) -> bool: ...


@tp.overload
def is_backprop_required(context: tp.Optional[Context]=None) -> bool: ...


# chainerx_cc/chainerx/python/backward.cc
def backward(
        arg0: tp.Union[ndarray, tp.List[ndarray]],
        backprop_id: tp.Optional[BackpropId]=None,
        enable_double_backprop: bool=...) -> None: ...


# python/check_backward.cc
def check_backward(
        func: tp.Callable[[ndarray], None],
        inputs: tp.List[ndarray],
        grad_outputs: tp.List[ndarray],
        eps: tp.List[ndarray],
        atol: float=...,
        rtol: float=...,
        backprop_id: tp.Optional[BackpropId]=None) -> None: ...


def check_double_backward(
        func: tp.Callable[[ndarray], None],
        inputs: tp.List[ndarray],
        grad_outputs: tp.List[ndarray],
        grad_grad_inputs: tp.List[ndarray],
        eps: tp.List[ndarray],
        atol: float=...,
        rtol: float=...,
        backprop_id: tp.Optional[BackpropId]=None) -> None: ...


# chainerx_cc/chainerx/python/context.cc
class Context:
    def get_backend(self, arg0: str) -> Backend: ...

    @tp.overload
    def get_device(self, arg0: str) -> Device: ...

    @tp.overload
    def get_device(self, arg0: str, arg1: int) -> Device: ...

    def make_backprop_id(self, backprop_name: str) -> BackpropId: ...

    def release_backprop_id(self, backprop_id: BackpropId) -> None: ...

    def _check_valid_backprop_id(self, backprop_id: BackpropId) -> None: ...


class ContextScope:
    def __enter__(self) -> None: ...

    def __exit__(self, *args) -> None: ...


def get_backend(arg0: str) -> Backend: ...


@tp.overload
def get_device(device: Device) -> Device: ...


@tp.overload
def get_device(arg0: str, arg1: tp.Optional[int]=None) -> Device: ...


def get_default_context() -> Context: ...


def set_default_context(arg0: Context) -> None: ...


def get_global_default_context() -> Context: ...


def set_global_default_context(arg0: Context) -> None: ...


# chainerx_cc/chainerx/python/device.cc
class Device:
    @property
    def backend(self) -> Backend: ...

    @property
    def context(self) -> Context: ...

    @property
    def index(self) -> int: ...

    @property
    def name(self) -> str: ...

    def synchronize(self) -> None: ...


def get_default_device() -> Device: ...


@tp.overload
def set_default_device(arg0: Device) -> None: ...


@tp.overload
def set_default_device(arg0: str) -> None: ...


@tp.overload
def using_device(arg0: Device) -> DeviceScope: ...


@tp.overload
def using_device(arg0: str, arg1: tp.Optional[int]=None) -> DeviceScope: ...


class DeviceScope:
    def __enter__(self) -> DeviceScope: ...

    def __exit__(self, *args) -> None: ...

    @property
    def device(self) -> Device: ...


# chainerx_cc/chainerx/python/error.cc
class BackendError(Exception): ...


class ChainerxError(Exception): ...


class ContextError(Exception): ...


class DeviceError(Exception): ...


class DimensionError(Exception): ...


class DtypeError(Exception): ...


class GradientCheckError(Exception): ...


class NotImplementedError(Exception): ...


# chainerx_cc/chainerx/python/graph.cc
class AnyGraph: ...


class BackpropId:
    @property
    def context(self) -> Context: ...

    @property
    def name(self) -> str: ...


class BackpropScope:
    def __enter__(self) -> BackpropId: ...

    def __exit__(self, *args) -> None: ...


# chainerx_cc/chainerx/python/array.cc
class ndarray:
    @property
    def T(self) -> ndarray: ...

    @property
    def data_ptr(self) -> int: ...

    @property
    def data_size(self) -> int: ...

    @property
    def device(self) -> Device: ...

    @property
    def dtype(self) -> numpy.dtype: ...

    @property
    def grad(self) -> tp.Optional[ndarray]: ...

    @property
    def is_contiguous(self) -> bool: ...

    @property
    def itemsize(self) -> int: ...

    @property
    def nbytes(self) -> int: ...

    @property
    def ndim(self) -> int: ...

    @property
    def offset(self) -> int: ...

    @property
    def shape(self) -> tp.Tuple[int, ...]: ...

    @property
    def size(self) -> int: ...

    @property
    def strides(self) -> tp.Tuple[int, ...]: ...

    def __add__(self, arg0: tp.Any) -> ndarray: ...

    def __and__(self, arg0: tp.Any) -> ndarray: ...

    def __bool__(self) -> bool: ...

    def __float__(self) -> float: ...

    def __ge__(self, arg0: ndarray) -> ndarray: ...

    def __getitem__(self, key: tp.Any) -> ndarray: ...

    def __gt__(self, arg0: tp.Any) -> ndarray: ...

    def __iadd__(self, arg0: tp.Any) -> ndarray: ...

    def __iand__(self, arg0: tp.Any) -> ndarray: ...

    def __ilshift__(self, arg0: tp.Any) -> ndarray: ...
    
    def __imod__(self, arg0: tp.Any) -> ndarray: ...

    def __imul__(self, arg0: tp.Any) -> ndarray: ...

    def __init__(
            self,
            shape: tp.Union[int, tp.Sequence[int]],
            dtype: tp.Any,
            device: tp.Optional[Device]=None) -> None: ...

    def __int__(self) -> int: ...

    def __ior__(self, arg0: tp.Any) -> ndarray: ...

    def __irshift__(self, arg0: tp.Any) -> ndarray: ...

    def __isub__(self, arg0: tp.Any) -> ndarray: ...

    def __itruediv__(self, arg0: tp.Any) -> ndarray: ...

    def __ixor__(self, arg0: tp.Any) -> ndarray: ...

    def __le__(self, arg0: ndarray) -> ndarray: ...

    def __len__(self) -> int: ...

    def __lshift__(self, arg0: tp.Any) -> ndarray: ...

    def __lt__(self, arg0: tp.Any) -> ndarray: ...

    def __mod__(self, arg0: tp.Any) -> ndarray: ...

    def __mul__(self, arg0: tp.Any) -> ndarray: ...

    def __neg__(self) -> ndarray: ...

    def __or__(self, arg0: tp.Any) -> ndarray: ...

    def __pow__(self, arg0: tp.Any) -> ndarray: ...

    def __radd__(self, arg0: tp.Any) -> ndarray: ...

    def __rand__(self, arg0: tp.Any) -> ndarray: ...

    def __repr__(self) -> str: ...

    def __rlshift__(self, arg0: tp.Any) -> ndarray: ...
    
    def __rmod__(self, arg0: tp.Any) -> ndarray: ...

    def __rmul__(self, arg0: tp.Any) -> ndarray: ...

    def __ror__(self, arg0: tp.Any) -> ndarray: ...

    def __rpow__(self, arg0: tp.Any) -> ndarray: ...

    def __rrshift__(self, arg0: tp.Any) -> ndarray: ...

    def __rshift__(self, arg0: tp.Any) -> ndarray: ...

    def __rsub__(self, arg0: tp.Any) -> ndarray: ...

    def __rxor__(self, arg0: tp.Any) -> ndarray: ...

    def __sub__(self, arg0: tp.Any) -> ndarray: ...

    def __truediv__(self, arg0: tp.Any) -> ndarray: ...

    def __xor__(self, arg0: tp.Any) -> ndarray: ...

    def argmax(self, axis: tp.Optional[int]=None) -> ndarray: ...

    def argmin(self, axis: tp.Optional[int]=None) -> ndarray: ...

    def flatten(self) -> ndarray: ...

    def ravel(self) -> ndarray: ...

    @tp.overload
    def as_grad_stopped(self, copy: bool=...) -> ndarray: ...

    @tp.overload
    def as_grad_stopped(
            self,
            arg0: tp.List[BackpropId],
            copy: bool=...) -> ndarray: ...

    def astype(self, dtype: tp.Any, copy: bool=...) -> ndarray: ...

    def backward(
            self,
            backprop_id: tp.Optional[BackpropId]=None,
            enable_double_backprop: bool=...) -> None: ...

    def cleargrad(self, backprop_id: tp.Optional[BackpropId]=None) -> None: ...

    def clip(self, a_min: tp.Optional[int], a_max: tp.Optional[int]): ...

    def copy(self) -> ndarray: ...

    def dot(self, b: ndarray) -> ndarray: ...

    def fill(self, value: tp.Any) -> None: ...

    def get_grad(self, backprop_id: tp.Optional[BackpropId]=None) -> ndarray: ...

    @tp.overload
    def is_backprop_required(
            self,
            backprop_id:
            tp.Optional[BackpropId]=None) -> bool: ...

    @tp.overload
    def is_backprop_required(self, backprop_id: AnyGraph) -> bool: ...

    def is_grad_required(
            self,
            backprop_id: tp.Optional[BackpropId]=None) -> bool: ...

    @tp.overload
    def max(self,
            axis: int,
            keepdims: bool=...) -> ndarray: ...

    @tp.overload
    def max(self,
            axis: tp.Optional[tp.Tuple[int, ...]]=None,
            keepdims: bool=...) -> ndarray: ...

    @tp.overload
    def min(self,
            axis: int,
            keepdims: bool=...) -> ndarray: ...

    @tp.overload
    def min(self,
            axis: tp.Optional[tp.Tuple[int, ...]]=None,
            keepdims: bool=...) -> ndarray: ...

    def ravel(self) -> ndarray: ...

    def require_grad(
            self,
            backprop_id: tp.Optional[BackpropId]=None) -> ndarray: ...

    @tp.overload
    def reshape(self, arg0: tp.Union[int, tp.Sequence[int]]) -> ndarray: ...

    @tp.overload
    def reshape(self, *args: tp.Any) -> ndarray: ...

    def set_grad(
            self,
            grad: tp.Optional[ndarray],
            backprop_id: tp.Optional[BackpropId]=None) -> None: ...

    @tp.overload
    def squeeze(self, axis: tp.Optional[tp.Tuple[int, ...]]=None) -> ndarray: ...

    @tp.overload
    def squeeze(self, axis: int) -> ndarray: ...

    def swapaxes(self, axis1: int, axis2: int) -> ndarray: ...

    @tp.overload
    def repeat(self, repeats: int, axis: tp.Optional[int]=None) -> ndarray: ...

    @tp.overload
    def repeat(self, repeats: tp.Tuple[int, ...], axis: tp.Optional[int]=None) -> ndarray: ...

    @tp.overload
    def sum(self, axis: int, keepdims: bool=...) -> ndarray: ...

    @tp.overload
    def sum(self,
            axis: tp.Optional[tp.Tuple[int, ...]]=None,
            keepdims: bool=...) -> ndarray: ...

    def take(self, indices: tp.Union[tp.Sequence[int], numpy.ndarray, ndarray],
             axis: tp.Optional[int]=None) -> ndarray: ...

    @tp.overload
    def to_device(self, arg0: Device) -> ndarray: ...

    @tp.overload
    def to_device(self, arg0: str, arg1: int) -> ndarray: ...

    @tp.overload
    def transpose(
            self,
            axes: tp.Optional[tp.List[int]]=None) -> ndarray: ...

    @tp.overload
    def transpose(self, *args: tp.Any) -> ndarray: ...

    @tp.overload
    def transpose(self, axes: int) -> ndarray: ...

    def view(self) -> ndarray: ...


# chainerx_cc/chainerx/python/routines.cc
def abs(x: ndarray) -> ndarray: ...


def absolute_error(x1: ndarray, x2: ndarray) -> ndarray: ...


def accuracy(y: ndarray, t: ndarray, ignore_label: tp.Optional[int]=None) -> ndarray: ...


def add(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def all(x: ndarray) -> ndarray: ...


def any(x: ndarray) -> ndarray: ...


def amax(a: ndarray,
         axis: tp.Union[int, tp.Optional[tp.List[int]]]=None,
         keepdims: bool=...) -> ndarray: ...


def amin(a: ndarray,
         axis: tp.Union[int, tp.Optional[tp.List[int]]]=None,
         keepdims: bool=...) -> ndarray: ...


def arange(
        start: tp.Any,
        stop: tp.Optional[tp.Any]=None,
        step: tp.Optional[tp.Any]=None,
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def arccos(x: ndarray) -> ndarray: ...


def arccosh(x: ndarray) -> ndarray: ...


def arcsin(x: ndarray) -> ndarray: ...


def arcsinh(x: ndarray) -> ndarray: ...


def arctan(x: ndarray) -> ndarray: ...


def arctan2(x1: ndarray, x2: ndarray) -> ndarray: ...


def argmax(a: ndarray, axis: tp.Optional[int]=None) -> ndarray: ...

def argmin(a: ndarray, axis: tp.Optional[int]=None) -> ndarray: ...

def nanargmax(a:ndarray, axis:tp.Optional[int]=None) -> ndarray: ...

def nanargmin(a:ndarray, axis:tp.Optional[int]=None) -> ndarray: ...


def array(
        object: tp.Any,
        dtype: tp.Optional[tp.Any]=None,
        copy: bool=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


def asarray(
        object: tp.Any,
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def ascontiguousarray(
        a: tp.Any,
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def atleast_2d(x: ndarray) -> ndarray: ...


def atleast_3d(x: ndarray) -> ndarray: ...


def average_pool(
        x: ndarray,
        ksize: tp.Union[int, tp.Tuple[int, ...]],
        stride: tp.Optional[tp.Union[int, tp.Tuple[int, ...]]]=None,
        pad: tp.Union[int, tp.Tuple[int, ...]]=...,
        pad_mode: str=...) -> ndarray: ...


def backprop_scope(backprop_name: str, context: tp.Any=None) -> BackpropScope: ...


def batch_norm(
        x: ndarray,
        gamma: ndarray,
        beta: ndarray,
        running_mean: ndarray,
        running_var: ndarray,
        eps: float=...,
        decay: float=...,
        axis: tp.Optional[tp.Union[int, tp.Tuple[int, ...]]]=None) -> ndarray: ...


def bitwise_and(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def bitwise_or(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def bitwise_xor(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def left_shift(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def right_shift(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def broadcast_to(array: ndarray, shape: tp.Tuple[int, ...]) -> ndarray: ...


def ceil(x: ndarray) -> ndarray: ...


def concatenate(arrays: tp.List[ndarray], axis: tp.Optional[int]=...) -> ndarray: ...


def context_scope(arg0: Context) -> ContextScope: ...


def conv(
        x: ndarray,
        w: ndarray,
        b: tp.Optional[ndarray]=None,
        stride: tp.Union[int, tp.Tuple[int, ...]]=...,
        pad: tp.Union[int, tp.Tuple[int, ...]]=...,
        cover_all: bool=False) -> ndarray: ...


def conv_transpose(
        x: ndarray,
        w: ndarray,
        b: tp.Optional[ndarray]=None,
        stride: tp.Union[int, tp.Tuple[int, ...]]=...,
        pad: tp.Union[int, tp.Tuple[int, ...]]=...,
        outsize: tp.Optional[tp.Tuple[int, ...]]=None) -> ndarray: ...


def copy(a: ndarray) -> ndarray: ...

def cos(x: ndarray) -> ndarray: ...

def cosh(x: ndarray) -> ndarray: ...

def diag(v: ndarray, k: int=..., device: tp.Optional[Device]=None) -> ndarray: ...


def diagflat(
        v: ndarray,
        k: int=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


def divide(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def dot(a: ndarray, b: ndarray) -> ndarray: ...


def dstack(arrays: tp.List[ndarray]) -> ndarray: ...


def empty(
        shape: tp.Union[int, tp.Tuple[int, ...]],
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def empty_like(a: ndarray, device: tp.Optional[Device]=None) -> ndarray: ...


def equal(x1: ndarray, x2: ndarray) -> ndarray: ...


def exp(x: ndarray) -> ndarray: ...


def expm1(x:ndarray) -> ndarray: ...


def exp2(x:ndarray) -> ndarray: ...


def eye(N: int,
        M: tp.Optional[int]=None,
        k: int=...,
        dtype: tp.Optional[tp.Any]=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


def expand_dims(a: ndarray, axis: int) -> ndarray: ...


def fabs(x: ndarray) -> ndarray: ...


def fixed_batch_norm(
        x: ndarray,
        gamma: ndarray,
        beta: ndarray,
        mean: ndarray,
        var: ndarray,
        eps: float=...,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None) -> ndarray: ...


def flip(
        a: ndarray,
        axes: tp.Optional[tp.Union[int, tp.List[int]]]=None) -> ndarray: ...


def fliplr(a: ndarray) -> ndarray: ...


def flipud(a: ndarray) -> ndarray: ...


def floor(x: ndarray) -> ndarray: ...


def fmod(x: ndarray) -> ndarray: ...


def frombuffer(
        buffer: tp.Any,
        dtype: tp.Optional[tp.Any]=...,
        count: int=...,
        offset: int=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


def full(
        shape: tp.Union[int, tp.Tuple[int, ...]],
        fill_value: tp.Any,
        dtype: tp.Optional[tp.Any],
        device: tp.Optional[Device]=None) -> ndarray: ...


def full_like(
        a: ndarray,
        fill_value: tp.Any,
        device: tp.Optional[Device]=None) -> ndarray: ...


def greater(x1: ndarray, x2: ndarray) -> ndarray: ...


def greater_equal(x1: ndarray, x2: ndarray) -> ndarray: ...

def gaussian_kl_divergence(mean: ndarray, ln_var: ndarray, reduce: tp.Optional[str]="sum") -> ndarray: ...

def hinge(x1: ndarray, x2: ndarray, norm: float=1.0) -> ndarray: ...

def hstack(arrays: tp.List[ndarray]) -> ndarray: ...

def huber_loss(x: ndarray, t: ndarray, delta: float,reduce: tp.Optional[str]="sum_along_second_axis") -> ndarray: ...

def identity(
        n: int,
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def is_available(): ...


def isfinite(x: ndarray) -> ndarray: ...


def isinf(x: ndarray) -> ndarray: ...


def isnan(x: ndarray) -> ndarray: ...


def less(x1: ndarray, x2: ndarray) -> ndarray: ...


def less_equal(x1: ndarray, x2: ndarray) -> ndarray: ...


def linear(
        x: ndarray,
        w: ndarray,
        b: tp.Optional[ndarray]=None,
        n_batch_axes: int=1) -> ndarray: ...

def lstm(c:ndarray, x:ndarray) -> tp.Tuple[ndarray, ...]: ...

def n_step_lstm(
    n_layers: int, 
    hx: ndarray,
    cx: ndarray,
    ws: tp.any,
    bs: tp.any,
    xs: tp.List[ndarray]) -> tp.Tuple[tp.List[ndarray], ...]: ...

def linspace(
        start: tp.Any,
        stop: tp.Any,
        num: int=...,
        endpoint: bool=True,
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def log(x: ndarray) -> ndarray: ...


def log10(x: ndarray) -> ndarray: ...


def log2(x: ndarray) -> ndarray: ...


def log1p(x: ndarray) -> ndarray: ...


def log_softmax(
        x: ndarray,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None) -> ndarray: ...


def logical_and(x1: ndarray, x2: ndarray) -> ndarray: ...


def logical_or(x1: ndarray, x2: ndarray) -> ndarray: ...

def logical_xor(x1: ndarray, x2: ndarray) -> ndarray: ...

def logical_not(x: ndarray) -> ndarray: ...


def logsumexp(
        x: ndarray,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None,
        keepdims: bool=...) -> ndarray: ...


def max_pool(
        x: ndarray,
        ksize: tp.Any,
        stride: tp.Union[int, tp.Tuple[int, ...]]=None,
        pad: tp.Union[int, tp.Tuple[int, ...]]=...,
        cover_all: bool=...) -> ndarray: ...


def maximum(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def meshgrid(arrays: tp.List[ndarray],
             indexing: tp.Optional[str]=...) -> tp.List[ndarray]: ...


def minimum(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def mod(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def moveaxis(a: ndarray, source: tp.Union[int, tp.Tuple[int, ...]],
             destination: tp.Union[int, tp.Tuple[int, ...]]) -> ndarray: ...


def multiply(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def nansum(a: ndarray,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None,
        keepdims: bool=...) -> ndarray: ...


def negative(x: ndarray) -> ndarray: ...


def not_equal(x1: ndarray, x2: ndarray) -> ndarray: ...


def nonzero(a: ndarray) -> ndarray: ...


def ones(shape: tp.Union[int, tp.Sequence[int]],
         dtype: tp.Optional[tp.Any]=None,
         device: tp.Optional[Device]=None) -> ndarray: ...


def ones_like(a: ndarray, device: tp.Optional[Device]=None) -> ndarray: ...


def remainder(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def reshape(
        a: ndarray,
        newshape: tp.Union[int, tp.Sequence[int]]) -> ndarray: ...


@tp.overload
def reshape(a: ndarray, *args: tp.Any) -> ndarray: ...

def sign(x: ndarray) -> ndarray: ...

def sin(x: ndarray) -> ndarray: ...

def sinh(x: ndarray) -> ndarray: ...

def sigmoid(x: ndarray) -> ndarray: ...

def sigmoid_cross_entropy(
        x1: ndarray,
        x2: ndarray) -> ndarray: ...

def relu(x: ndarray) -> ndarray: ...

def softmax(
        x: ndarray,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None) -> ndarray: ...

def softmax_cross_entropy(
        x1: ndarray,
        x2: ndarray) -> ndarray: ...

def softplus(x: ndarray, beta: double=1.0) -> ndarray: ...

def split(
        ary: ndarray,
        indices_or_sections: tp.Union[int, tp.List[int]],
        axis: int=...) -> tp.List[ndarray]: ...

def dsplit(
        ary: ndarray,
        indices_or_sections: tp.Union[int, tp.List[int]],
        ) -> tp.List[ndarray]: ...

def vsplit(
        ary: ndarray,
        indices_or_sections: tp.Union[int, tp.List[int]],
        ) -> tp.List[ndarray]: ...

def hsplit(
        ary: ndarray,
        indices_or_sections: tp.Union[int, tp.List[int]],
        ) -> tp.List[ndarray]: ...


def square(x: ndarray) -> ndarray: ...


def squared_error(x1: ndarray, x2: ndarray) -> ndarray: ...


def squared_difference(x1: ndarray, x2: ndarray) -> ndarray: ...


def sqrt(x: ndarray) -> ndarray: ...

def power(x1: tp.Any, x2: tp.Any) -> ndarray: ...

def squeeze(
        a: ndarray,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None) -> ndarray: ...


def stack(arrays: tp.List[ndarray], axis: int=...) -> ndarray: ...


def subtract(x1: tp.Any, x2: tp.Any) -> ndarray: ...


def sum(a: ndarray,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None,
        keepdims: bool=...) -> ndarray: ...


def count_nonzero(
        a: ndarray,
        axis: tp.Optional[tp.Union[int, tp.List[int]]]=None) -> ndarray: ...


def cumsum(
        x: ndarray,
        axis: tp.Optional[int]=None) -> ndarray: ...


def swapaxes(a: ndarray, axis1: int, axis2: int) -> ndarray: ...

@tp.overload
def repeat(a: ndarray, repeats: int, axis: tp.Optional[int]=None) -> ndarray: ...

@tp.overload
def repeat(a: ndarray, repeats: tp.Tuple[int, ...], axis: tp.Optional[int]=None) -> ndarray: ...


def take(a: ndarray, indices: ndarray, axis: tp.Optional[int]) -> ndarray: ...


def tan(x: ndarray) -> ndarray: ...


def tanh(x: ndarray) -> ndarray: ...


def to_numpy(array: ndarray, copy: bool=...) -> numpy.ndarray: ...


def _to_cupy(array: ndarray) -> numpy.ndarray: ...


def transpose(
        a: ndarray,
        axes: tp.Optional[tp.Union[int, tp.List[int]]]=None) -> ndarray: ...


def tri(N: int,
        M: tp.Optional[int]=None,
        k: int=...,
        dtype: tp.Optional[tp.Any]=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


def tril(m: ndarray, k: int=...) -> ndarray: ...


def triu(m: ndarray, k: int=...) -> ndarray: ...


def vstack(arrays: tp.List[ndarray]) -> ndarray: ...

def where(cond: ndarray, x: ndarray, y: ndarray) -> ndarray: ...


def zeros(
        shape: tp.Union[int, tp.Sequence[int]],
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def zeros_like(a: ndarray, device: tp.Optional[Device]=None) -> ndarray: ...


# chainerx_cc/chainerx/python/chainer_interop.cc
def _function_node_forward(
        function_node: function_node.FunctionNode,
        inputs: tp.List[ndarray],
        outputs: tp.List[ndarray],
        input_indexes_to_retain: tp.List[int],
        output_indexes_to_retain: tp.List[int]) -> None: ...


# chainerx/creation/from_data.py
def asanyarray(
        a: ndarray,
        dtype: tp.Optional[tp.Any]=None,
        device: tp.Optional[Device]=None) -> ndarray: ...


def fromfile(
        file: str,
        dtype: tp.Optional[tp.Any]=...,
        count: int=...,
        sep: str=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


def fromfunction(
        function: tp.Callable[..., tp.Any],
        shape: tp.Union[int, tp.Sequence[int]],
        **kwargs: tp.Any) -> ndarray: ...


def fromiter(
        iterable: tp.Iterable[tp.Any],
        dtype: tp.Optional[tp.Any],
        count: int=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


def fromstring(
        string: str,
        dtype: tp.Optional[tp.Any]=float,
        count=...,
        sep=...,
        device=None) -> ndarray: ...


def loadtxt(
        fname: str,
        dtype: tp.Optional[tp.Any]=...,
        comments: tp.Optional[tp.Union[str, tp.Iterable[str]]]=...,
        delimiter: tp.Optional[str]=None,
        converters: tp.Optional[tp.Dict[int, tp.Callable[[str], tp.Any]]]=None,
        skiprows: int=...,
        usecols: tp.Optional[tp.Union[int, tp.Iterable[int]]]=None,
        unpack: bool=...,
        ndmin: int=...,
        encoding: tp.Optional[str]=...,
        device: tp.Optional[Device]=None) -> ndarray: ...


# chainerx/math/misc.py
def clip(a: ndarray, a_min: tp.Any, a_max: tp.Any) -> ndarray: ...



import warnings


class IntervalTrigger(object):

    """Trigger based on a fixed interval.

    This trigger accepts iterations divided by a given interval. There are two
    ways to specify the interval: per iterations and epochs. `Iteration` means
    the number of updates, while `epoch` means the number of sweeps over the
    training dataset. Fractional values are allowed if the interval is a
    number of epochs; the trigger uses the `iteration` and `epoch_detail`
    attributes defined by the updater.

    For the description of triggers, see :func:`~chainer.training.get_trigger`.

    Args:
        period (int or float): Length of the interval. Must be an integer if
            unit is ``'iteration'``.
        unit (str): Unit of the length specified by ``period``. It must be
            either ``'iteration'`` or ``'epoch'``.

    """

    def __init__(self, period, unit):
        if unit not in ('epoch', 'iteration'):
            raise ValueError(
                'Trigger unit must be either \'epoch\' or \'iteration\'.')

        self.period = period
        self.unit = unit

        self._previous_iteration = 0
        self._previous_epoch_detail = 0.

        # count is kept for backward compatibility
        self.count = 0

    def __call__(self, trainer):
        """Decides whether the extension should be called on this iteration.

        Args:
            trainer (Trainer): Trainer object that this trigger is associated
                with. The updater associated with this trainer is used to
                determine if the trigger should fire.

        Returns:
            bool: True if the corresponding extension should be invoked in this
            iteration.

        """
        updater = trainer.updater
        if self.unit == 'epoch':
            epoch_detail = updater.epoch_detail
            previous_epoch_detail = self._previous_epoch_detail

            # if previous_epoch_detail is invalid value,
            # use the value of updater.
            if previous_epoch_detail < 0:
                previous_epoch_detail = updater.previous_epoch_detail

            # count is kept for backward compatibility
            self.count = epoch_detail // self.period

            fire = previous_epoch_detail // self.period != \
                epoch_detail // self.period
        else:
            iteration = updater.iteration
            previous_iteration = self._previous_iteration

            # if previous_iteration is invalid value,
            # guess it from current iteration.
            if previous_iteration < 0:
                previous_iteration = iteration - 1

            fire = previous_iteration // self.period != \
                iteration // self.period

        # save current values
        self._previous_iteration = updater.iteration
        if hasattr(updater, 'epoch_detail'):
            self._previous_epoch_detail = updater.epoch_detail

        return fire

    def serialize(self, serializer):
        try:
            self._previous_iteration = serializer(
                'previous_iteration', self._previous_iteration)
        except KeyError:
            warnings.warn(
                'The previous value of iteration is not saved. '
                'IntervalTrigger guesses it using current iteration. '
                'If this trigger is not called at every iteration, '
                'it may not work correctly.')
            # set a negative value for invalid
            self._previous_iteration = -1

        try:
            self._previous_epoch_detail = serializer(
                'previous_epoch_detail', self._previous_epoch_detail)
        except KeyError:
            warnings.warn(
                'The previous value of epoch_detail is not saved. '
                'IntervalTrigger uses the value of '
                'trainer.updater.previous_epoch_detail. '
                'If this trigger is not called at every iteration, '
                'it may not work correctly.')
            # set a negative value for invalid
            self._previous_epoch_detail = -1.

    def get_training_length(self):
        return (self.period, self.unit)

    def __str__(self):
        """Returns a string describing the class and interval

        Returns:
            str: IntervalTrigger(<period>, '<unit>')
        """
        return '{}({}, \'{}\')'.format(
            self.__class__.__name__, self.period, self.unit
        )



import numpy as np
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE

from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


@support((1, 6, 7))
def convert_Add(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Add', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Add', input_names, output_names),


@support((1, 6, 7))
def convert_AddConstant(
        func, opset_version, input_names, output_names, context):
    value_name = context.add_const(
        np.array(func.value, dtype=func.inputs[0].dtype), 'value')
    input_names.append(value_name)

    if opset_version == 1:
        return onnx_helper.make_node(
            'Add', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Add', input_names, output_names),


@support((1, 6, 7))
def convert_Sub(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Sub', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Sub', input_names, output_names),


@support((1, 6, 7))
def convert_SubFromConstant(
        func, opset_version, input_names, output_names, context):
    value_name = context.add_const(
        np.array(func.value, dtype=func.inputs[0].dtype), 'value')
    input_names[:0] = [value_name]

    if opset_version == 1:
        return onnx_helper.make_node(
            'Sub', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Sub', input_names, output_names),


@support((1, 6, 7))
def convert_Mul(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Mul', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Mul', input_names, output_names),


@support((1, 6, 7))
def convert_MulConstant(
        func, opset_version, input_names, output_names, context):
    value_name = context.add_const(
        np.array(func.value, dtype=func.inputs[0].dtype), 'value')
    input_names.append(value_name)

    if opset_version == 1:
        return onnx_helper.make_node(
            'Mul', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Mul', input_names, output_names),


@support((1, 6))
def convert_Neg(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Neg', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6:
        return onnx_helper.make_node('Neg', input_names, output_names),


@support((1, 6, 7))
def convert_Div(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Div', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Div', input_names, output_names),


@support((1, 6, 7))
def convert_DivFromConstant(
        func, opset_version, input_names, output_names, context):
    value_name = context.add_const(
        np.array(func.value, dtype=func.inputs[0].dtype), 'value')
    input_names[:0] = [value_name]

    if opset_version == 1:
        return onnx_helper.make_node(
            'Div', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node('Div', input_names, output_names),


@support((1, 6))
def convert_Absolute(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Abs', input_names, output_names, consumed_inputs=[1]),
    elif opset_version == 6:
        return onnx_helper.make_node('Abs', input_names, output_names),


@support((7,))
def convert_Arccos(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Acos', input_names, output_names),


@support((7,))
def convert_Arcsin(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Asin', input_names, output_names),


@support((7,))
def convert_Arctan(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Atan', input_names, output_names),


@support((1, 7))
def convert_PowConstVar(
        func, opset_version, input_names, output_names, context):
    value_name = context.add_const(
        np.array(func.value, dtype=func.inputs[0].dtype), 'value')
    input_names.insert(0, value_name)

    if opset_version == 1 or opset_version == 7:
        return onnx_helper.make_node('Pow', input_names, output_names),


@support((1, 7))
def convert_PowVarConst(
        func, opset_version, input_names, output_names, context):
    value_name = context.add_const(
        np.array(func.value, dtype=func.inputs[0].dtype), 'value')
    input_names.append(value_name)

    if opset_version == 1 or opset_version == 7:
        return onnx_helper.make_node('Pow', input_names, output_names),


@support((1, 7))
def convert_PowVarVar(
        func, opset_version, input_names, output_names, context):
    if opset_version == 1 or opset_version == 7:
        return onnx_helper.make_node('Pow', input_names, output_names),


@support((1, 6))
def convert_Clip(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Clip', input_names, output_names,
            max=func.x_max,
            min=func.x_min,
            consumed_inputs=[1]
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'Clip', input_names, output_names,
            max=func.x_max,
            min=func.x_min,
        ),


@support((7,))
def convert_Cos(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Cos', input_names, output_names),


@support((9,))
def convert_Cosh(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Cosh', input_names, output_names),


@support((1, 6))
def convert_Exp(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Exp', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6:
        return onnx_helper.make_node('Exp', input_names, output_names),


def convert_Identity(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Identity', input_names, output_names),


def convert_MatMul(func, opset_version, input_names, output_names, context):
    ndim_a = len(func.inputs[0].shape)
    ndim_b = len(func.inputs[1].shape)

    gb = onnx_helper.GraphBuilder()
    if ndim_a > 1 and func.transa:
        perm = list(range(ndim_a))
        perm[-1], perm[-2] = perm[-2], perm[-1]
        input_names[0] = gb.op('Transpose', [input_names[0]], perm=perm)
    if ndim_b > 1 and func.transb:
        perm = list(range(ndim_b))
        perm[-1], perm[-2] = perm[-2], perm[-1]
        input_names[1] = gb.op('Transpose', [input_names[1]], perm=perm)
    gb.op('MatMul', input_names)
    return gb.nodes(output_names)


@support((1, 6, 8))
def convert_Maximum(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Max', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 8:
        return onnx_helper.make_node('Max', input_names, output_names),


@support((1, 6, 8))
def convert_Minimum(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Min', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 8:
        return onnx_helper.make_node('Min', input_names, output_names),


@support((7,))
def convert_Sin(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Sin', input_names, output_names),


@support((9,))
def convert_Sinh(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Sinh', input_names, output_names),


@support((1, 6))
def convert_Sqrt(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Sqrt', input_names, output_names, consumed_inputs=[1, 1]),
    elif opset_version == 6:
        return onnx_helper.make_node('Sqrt', input_names, output_names),


def convert_RsqrtGPU(func, opset_version, input_names, output_names, context):
    gb = onnx_helper.GraphBuilder()
    sqrt_out = gb.op('Sqrt', input_names)
    gb.op('Reciprocal', [sqrt_out])
    return gb.nodes(output_names)


@support((6,))
def convert_Log(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Log', input_names, output_names),


def convert_LogSumExp(func, opset_version, input_names, output_names, context):
    # Use keepdims=False by default
    # since the chainer does not support keepdims option
    kwargs = {'keepdims': False}
    if hasattr(func, 'keepdims'):
        kwargs['keepdims'] = func.keepdims
    if func.axis is not None:
        kwargs['axes'] = func.axis
    return onnx_helper.make_node(
        'ReduceLogSumExp', input_names, output_names, **kwargs),


def convert_Max(func, opset_version, input_names, output_names, context):
    kwargs = {'keepdims': func.keepdims}
    if func.axis is not None:
        kwargs['axes'] = func.axis
    return onnx_helper.make_node(
        'ReduceMax', input_names, output_names, **kwargs),


def convert_Mean(func, opset_version, input_names, output_names, context):
    kwargs = {'keepdims': func.keepdims}
    if func.axis is not None:
        kwargs['axes'] = func.axis
    return onnx_helper.make_node(
        'ReduceMean', input_names, output_names, **kwargs),


def convert_Min(func, opset_version, input_names, output_names, context):
    kwargs = {'keepdims': func.keepdims}
    if func.axis is not None:
        kwargs['axes'] = func.axis
    return onnx_helper.make_node(
        'ReduceMin', input_names, output_names, **kwargs),


def convert_Prod(func, opset_version, input_names, output_names, context):
    kwargs = {'keepdims': func.keepdims}
    if func.axis is not None:
        kwargs['axes'] = func.axis
    return onnx_helper.make_node(
        'ReduceProd', input_names, output_names, **kwargs),


def convert_Sum(func, opset_version, input_names, output_names, context):
    kwargs = {'keepdims': func.keepdims}
    if func.axis is not None:
        kwargs['axes'] = func.axis
    return onnx_helper.make_node(
        'ReduceSum', input_names, output_names, **kwargs),


@support((7,))
def convert_Tan(func, opset_version, input_names, output_names, context):
    return onnx_helper.make_node('Tan', input_names, output_names),


@support((1, 6, 7))
def convert_LinearInterpolate(
        func, opset_version, input_names, output_names, context):
    typ = func.inputs[0].dtype if isinstance(
        func.inputs[0].dtype, np.dtype) else np.dtype(func.inputs[0].dtype)

    one_name = context.add_const(np.array(1, dtype=typ), 'one')

    kwargs = {'consumed_inputs': [1, 1]} if opset_version == 1 else {}
    kwargs2 = {} if opset_version >= 7 else {'broadcast': 1}

    gb = onnx_helper.GraphBuilder()
    p, x, y = input_names
    n1 = gb.op('Sub', [one_name, p], **kwargs, **kwargs2)
    n2 = gb.op('Mul', [p, x], **kwargs)
    n3 = gb.op('Mul', [n1, y], **kwargs)
    gb.op_output_named('Add', [n2, n3], output_names, **kwargs)

    return gb.nodes()


@support((1, 6, 7))
def convert_Square(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Mul', [input_names[0], input_names[0]], output_names,
            consumed_inputs=[1, 1]),
    elif opset_version == 6 or opset_version == 7:
        return onnx_helper.make_node(
            'Mul', [input_names[0], input_names[0]], output_names),


@support((8,))
def convert_BroadcastTo(
        func, opset_version, input_names, output_names, context):
    shape_name = context.add_const(
        np.array(func._shape, dtype=np.int64), 'shape')
    input_names.append(shape_name)
    return onnx_helper.make_node('Expand', input_names, output_names),


def _argminmax_nodes(op_name, func, input_names, output_names, context):
    gb = onnx_helper.GraphBuilder()
    target_input_names = input_names
    axis = func.axis
    if axis is None:
        shape_name = context.add_const(np.array([-1], dtype=np.int64), 'shape')
        input_names.append(shape_name)
        target_input_names = [gb.op('Reshape', input_names)]
        axis = 0
    out = gb.op(op_name, target_input_names, axis=axis, keepdims=0)
    # Chainer's ArgMax always return value as int32
    # Cast spec is changed from opset6, this logic does not support ~opset5
    gb.op('Cast', [out], to=NP_TYPE_TO_TENSOR_TYPE[np.dtype('int32')])
    return gb.nodes(output_names)


@support((6,))
def convert_ArgMax(func, opset_version, input_names, output_names, context):
    return _argminmax_nodes('ArgMax', func, input_names, output_names, context)


@support((6,))
def convert_ArgMin(func, opset_version, input_names, output_names, context):
    return _argminmax_nodes('ArgMin', func, input_names, output_names, context)



import numpy as np


def shaped_range(*shape, dtype=np.float32):
    r = np.arange(np.prod(shape))
    r = r.reshape(shape)
    return r


def _increasing_impl(*shape, dtype=np.float32, negative=True, bias=0):
    r = shaped_range(*shape, dtype=dtype)
    if negative:
        r -= r.size // 2
    if dtype in (np.float32, np.float64):
        r = r * 0.5
    r += bias
    return r.astype(dtype)


def increasing(*shape, dtype=np.float32):
    """Returns a monotonically increasing ndarray for test inputs.

    The output will contain both zero, negative numbers, and non
    integer numbers for float dtypes. A test writer is supposed to
    consider this function first.

    Example:

    >>> onnx_chainer.testing.input_generator.increasing(3, 4)
    array([[-3. , -2.5, -2. , -1.5],
           [-1. , -0.5,  0. ,  0.5],
           [ 1. ,  1.5,  2. ,  2.5]], dtype=float32)

    Args:
        shape (tuple of int): The shape of the output array.
        dtype (numpy.dtype): The dtype of the output array.

    Returns:
        numpy.ndarray
    """
    return _increasing_impl(*shape, dtype=dtype)


def nonzero_increasing(*shape, dtype=np.float32, bias=1e-7):
    """Returns a monotonically increasing ndarray for test inputs.

    Similar to `increasing` but contains no zeros. Expected to be used
    for divisors.

    Example:

    >>> onnx_chainer.testing.input_generator.nonzero_increasing(3, 4)
    array([[-3.0000000e+00, -2.5000000e+00, -1.9999999e+00, -1.4999999e+00],
           [-9.9999988e-01, -4.9999991e-01,  1.0000000e-07,  5.0000012e-01],
           [ 1.0000001e+00,  1.5000001e+00,  2.0000000e+00,  2.5000000e+00]],
          dtype=float32)

    Args:
        shape (tuple of int): The shape of the output array.
        dtype (numpy.dtype): The dtype of the output array.
        bias (float): The bias to avoid zero.

    Returns:
        numpy.ndarray
    """
    assert dtype in (np.float32, np.float64)
    return _increasing_impl(*shape, dtype=dtype, bias=bias)


def positive_increasing(*shape, dtype=np.float32, bias=1e-7):
    """Returns a monotonically increasing ndarray for test inputs.

    Similar to `increasing` but contains only positive numbers. Expected
    to be used for `math.log`, `math.sqrt`, etc.

    Example:

    >>> onnx_chainer.testing.input_generator.positive_increasing(3, 4)
    array([[1.0000000e-07, 5.0000012e-01, 1.0000001e+00, 1.5000001e+00],
           [2.0000000e+00, 2.5000000e+00, 3.0000000e+00, 3.5000000e+00],
           [4.0000000e+00, 4.5000000e+00, 5.0000000e+00, 5.5000000e+00]],
          dtype=float32)

    Args:
        shape (tuple of int): The shape of the output array.
        dtype (numpy.dtype): The dtype of the output array.
        bias (float): The bias to avoid zero.

    Returns:
        numpy.ndarray
    """
    return _increasing_impl(*shape, dtype=dtype, negative=False, bias=bias)



import numpy

from chainer import backend
from chainer import function_node
from chainer.utils import argument
from chainer.utils import type_check


# {numpy: True, cupy: False}
_xp_supports_batch_eigh = {}


# routines for batched matrices
def _eigh(a, xp):
    if xp not in _xp_supports_batch_eigh:
        try:
            xp.linalg.eigh(xp.ones((2, 2, 2), xp.float32))
        except ValueError:
            _xp_supports_batch_eigh[xp] = False
        else:
            _xp_supports_batch_eigh[xp] = True
    if _xp_supports_batch_eigh[xp]:
        return xp.linalg.eigh(a)
    ws = []
    vs = []
    for ai in a:
        w, v = xp.linalg.eigh(ai)
        ws.append(w)
        vs.append(v)
    return xp.stack(ws), xp.stack(vs)


def _matmul(a, b, xp):
    if hasattr(xp, 'matmul'):  # numpy.matmul is supported from version 1.10.0
        return xp.matmul(a, b)
    else:
        return xp.einsum('bij,bjk->bik', a, b)


def _diag(a, xp):
    s0, s1 = a.shape
    ret = xp.zeros((s0, s1, s1), a.dtype)
    arange_s1 = numpy.arange(s1)
    ret[:, arange_s1, arange_s1] = a
    return ret


def _calc_axis_and_m(x_shape, batch_size):
    m = batch_size
    spatial_ndim = len(x_shape) - 2
    spatial_axis = tuple(range(2, 2 + spatial_ndim))
    for i in spatial_axis:
        m *= x_shape[i]
    return spatial_axis, m


class DecorrelatedBatchNormalization(function_node.FunctionNode):

    def __init__(self, groups=16, eps=2e-5, mean=None, projection=None,
                 decay=0.9):
        self.groups = groups

        self.running_mean = mean
        self.running_projection = projection

        self.eps = eps
        self.decay = decay
        self.axis = None

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 1)
        x_type = in_types[0]
        type_check.expect(
            x_type.dtype.kind == 'f',
            x_type.shape[1] % self.groups == 0,
        )
        type_check.expect(
            x_type.ndim >= 2,
        )

    def forward(self, inputs):
        self.retain_inputs(())
        x = inputs[0]
        xp = backend.get_array_module(x)
        x_shape = x.shape
        b, c = x_shape[:2]
        g = self.groups
        C = c // g
        spatial_axis, m = _calc_axis_and_m(x_shape, b)

        # (g, C, m)
        x_hat = x.transpose((1, 0) + spatial_axis).reshape(g, C, m)
        mean = x_hat.mean(axis=2, keepdims=True)
        x_hat = x_hat - mean
        self.eps = x.dtype.type(self.eps)

        eps_matrix = self.eps * xp.eye(C, dtype=x.dtype)
        cov = _matmul(
            x_hat, x_hat.transpose(0, 2, 1),
            xp) / x.dtype.type(m) + eps_matrix
        # (g, C), (g, C, C)
        self.eigvals, self.eigvectors = _eigh(cov, xp)
        U = _matmul(
            _diag(self.eigvals ** -0.5, xp),
            self.eigvectors.transpose(0, 2, 1),
            xp)
        self.y_hat_pca = _matmul(U, x_hat, xp)  # PCA whitening
        # ZCA whitening
        y_hat = _matmul(self.eigvectors, self.y_hat_pca, xp)

        y = y_hat.reshape((c, b) + x_shape[2:]).transpose(
            (1, 0) + spatial_axis)

        # Update running statistics
        if self.running_mean is not None:
            mean = mean.squeeze(axis=2)
            self.running_mean *= self.decay
            self.running_mean += (1 - self.decay) * mean
        if self.running_projection is not None:
            adjust = m / max(m - 1., 1.)  # unbiased estimation
            self.running_projection *= self.decay
            projection = _matmul(self.eigvectors, U, xp)
            self.running_projection += (1 - self.decay) * adjust * projection

        return y,

    def backward(self, indexes, grad_outputs):
        gy, = grad_outputs

        f = DecorrelatedBatchNormalizationGrad(
            self.groups, self.eigvals, self.eigvectors, self.y_hat_pca)
        return f.apply((gy,))


class DecorrelatedBatchNormalizationGrad(function_node.FunctionNode):

    def __init__(self, groups, eigvals, eigvectors, y_hat_pca):
        self.groups = groups
        self.eigvals = eigvals
        self.eigvectors = eigvectors
        self.y_hat_pca = y_hat_pca

    def forward(self, inputs):
        self.retain_inputs(())
        gy = inputs[0]
        xp = backend.get_array_module(gy)
        gy_shape = gy.shape
        b, c = gy_shape[:2]
        g = self.groups
        C = c // g
        spatial_axis, m = _calc_axis_and_m(gy_shape, b)
        arange_C = numpy.arange(C)
        diag_indices = slice(None), arange_C, arange_C

        gy_hat = gy.transpose((1, 0) + spatial_axis).reshape(g, C, m)

        eigvectors = self.eigvectors
        eigvals = self.eigvals
        y_hat_pca = self.y_hat_pca
        gy_hat_pca = _matmul(eigvectors.transpose(0, 2, 1), gy_hat, xp)
        f = gy_hat_pca.mean(axis=2, keepdims=True)

        K = eigvals[:, :, None] - eigvals[:, None, :]
        valid = K != 0  # to avoid nan, use eig_i != eig_j instead of i != j
        K[valid] = xp.reciprocal(K[valid])

        V = _diag(eigvals, xp)
        V_sqrt = _diag(eigvals ** 0.5, xp)
        V_invsqrt = _diag(eigvals ** -0.5, xp)

        F_c = _matmul(
            gy_hat_pca, y_hat_pca.transpose(0, 2, 1),
            xp) / gy.dtype.type(m)
        M = xp.zeros_like(F_c)
        M[diag_indices] = F_c[diag_indices]

        mat = K.transpose(0, 2, 1) * (
            _matmul(V, F_c.transpose(0, 2, 1), xp)
            + _matmul(_matmul(V_sqrt, F_c, xp), V_sqrt, xp)
        )
        S = mat + mat.transpose(0, 2, 1)
        R = gy_hat_pca - f + _matmul(
            (S - M).transpose(0, 2, 1), y_hat_pca, xp)
        gx_hat = _matmul(
            _matmul(R.transpose(0, 2, 1), V_invsqrt, xp),
            eigvectors.transpose(0, 2, 1), xp
        ).transpose(0, 2, 1)

        gx = gx_hat.reshape((c, b) + gy_shape[2:]).transpose(
            (1, 0) + spatial_axis)

        self.retain_outputs(())
        return gx,

    def backward(self, inputs, grad_outputs):
        # TODO(crcrpar): Implement this.
        raise NotImplementedError('Double backward is not implemented for'
                                  ' decorrelated batch normalization.')


class FixedDecorrelatedBatchNormalization(function_node.FunctionNode):

    def __init__(self, groups):
        self.groups = groups

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 3)
        x_type, mean_type, var_type = in_types
        type_check.expect(
            x_type.dtype.kind == 'f',
            mean_type.dtype == x_type.dtype,
            var_type.dtype == x_type.dtype,
        )
        type_check.expect(
            x_type.ndim >= 2,
        )

    def forward(self, inputs):
        self.retain_inputs((0, 1, 2))
        x, mean, projection = inputs
        xp = backend.get_array_module(x)
        x_shape = x.shape
        b, c = x_shape[:2]
        g = self.groups
        C = c // g
        spatial_axis, m = _calc_axis_and_m(x_shape, b)

        x_hat = x.transpose((1, 0) + spatial_axis).reshape(g, C, m)
        x_hat = x_hat - xp.expand_dims(mean, axis=2)

        y_hat = _matmul(projection, x_hat, xp)

        y = y_hat.reshape((c, b) + x_shape[2:]).transpose(
            (1, 0) + spatial_axis)

        return y,

    def backward(self, indexes, grad_outputs):
        x, mean, projection = self.get_retained_inputs()
        gy,  = grad_outputs
        f = FixedDecorrelatedBatchNormalizationGrad(self.groups)
        return f.apply((x, mean, projection, gy))


class FixedDecorrelatedBatchNormalizationGrad(function_node.FunctionNode):

    def __init__(self, groups):
        self.groups = groups

    def forward(self, inputs):
        self.retain_inputs(())
        x, mean, projection, gy = inputs
        xp = backend.get_array_module(x)
        gy_shape = gy.shape
        b, c = gy_shape[:2]
        g = self.groups
        C = c // g
        spatial_axis, m = _calc_axis_and_m(gy_shape, b)

        gy_hat = gy.transpose((1, 0) + spatial_axis).reshape(g, C, m)
        x_hat = x.transpose((1, 0) + spatial_axis).reshape(g, C, m)
        gy_hat_pca = _matmul(projection.transpose(0, 2, 1), gy_hat, xp)
        gx = gy_hat_pca.reshape((c, b) + gy_shape[2:]).transpose(
            (1, 0) + spatial_axis)
        rhs = x_hat - xp.expand_dims(mean, axis=2)
        gprojection = _matmul((x_hat - rhs).transpose(0, 2, 1), gy_hat, xp)
        gmean = -gy_hat_pca[..., 0]
        self.retain_outputs(())
        return gx, gmean, gprojection

    def backward(self, inputs, grad_outputs):
        # TODO(crcrpar): Implement this.
        raise NotImplementedError('Double backward is not implemented for'
                                  ' fixed decorrelated batch normalization.')


def decorrelated_batch_normalization(x, **kwargs):
    """decorrelated_batch_normalization(x, *, groups=16, eps=2e-5, \
running_mean=None, running_projection=None, decay=0.9)

    Decorrelated batch normalization function.

    It takes the input variable ``x`` and normalizes it using
    batch statistics to make the output zero-mean and decorrelated.

    Args:
        x (:class:`~chainer.Variable`): Input variable.
        groups (int): Number of groups to use for group whitening.
        eps (float): Epsilon value for numerical stability.
        running_mean (:ref:`ndarray`): Expected value of the mean. This is a
            running average of the mean over several mini-batches using
            the decay parameter. If ``None``, the expected mean is initialized
            to zero.
        running_projection (:ref:`ndarray`):
            Expected value of the project matrix. This is a
            running average of the projection over several mini-batches using
            the decay parameter. If ``None``, the expected projected is
            initialized to the identity matrix.
        decay (float): Decay rate of moving average. It is used during
            training.

    Returns:
        ~chainer.Variable: The output variable which has the same shape as
        :math:`x`.

    See: `Decorrelated Batch Normalization <https://arxiv.org/abs/1804.08450>`_

    .. seealso:: :class:`~chainer.links.DecorrelatedBatchNormalization`

    """
    groups, eps, running_mean, running_projection, decay = \
        argument.parse_kwargs(
            kwargs, ('groups', 16), ('eps', 2e-5), ('running_mean', None),
            ('running_projection', None), ('decay', 0.9))

    f = DecorrelatedBatchNormalization(
        groups, eps, running_mean, running_projection, decay)
    return f.apply((x,))[0]


def fixed_decorrelated_batch_normalization(x, mean, projection, groups=16):
    """Decorrelated batch normalization function with fixed statistics.

    This is a variant of decorrelated batch normalization, where the mean and
    projection statistics are given by the caller as fixed variables. This is
    used in testing mode of the decorrelated batch normalization layer, where
    batch statistics cannot be used for prediction consistency.

    Args:
        x (:class:`~chainer.Variable`): Input variable.
        mean (:class:`~chainer.Variable` or :ref:`ndarray`):
            Shifting parameter of input.
        projection (:class:`~chainer.Variable` or :ref:`ndarray`):
            Projection matrix for decorrelation of input.
        groups (int): Number of groups to use for group whitening.

    Returns:
        ~chainer.Variable: The output variable which has the same shape as
        :math:`x`.

    .. seealso::
       :func:`~chainer.functions.decorrelated_batch_normalization`,
       :class:`~chainer.links.DecorrelatedBatchNormalization`

    """
    f = FixedDecorrelatedBatchNormalization(groups)
    return f.apply((x, mean, projection))[0]



from __future__ import print_function

from collections import OrderedDict
import warnings

import chainer

try:
    import onnx
    from onnx import checker
    from onnx import helper
    from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE
    from onnx import numpy_helper
    from onnx import shape_inference

    from onnx_chainer.context import Context
    from onnx_chainer.graph import Graph
    from onnx_chainer import mapping
    from onnx_chainer.onnx_helper import is_support_non_standard_domain

    _available = True
except ImportError:
    _available = False

MINIMUM_OPSET_VERSION = 7
MAXIMUM_OPSET_VERSION = 10


def _check_available():
    if not _available:
        raise ImportError(
            'ONNX is not installed on your environment. Exporting your model '
            'in ONNX format needs the onnx package.\n\n'
            '\t$ pip install \'onnx<1.6.0\'\n\n')


def convert_parameter(parameter, context):
    if isinstance(parameter, chainer.Parameter):
        array = parameter.array
    elif isinstance(parameter, chainer.Variable):
        array = parameter.array
    elif isinstance(parameter, chainer.get_array_types()):
        array = parameter
    else:
        raise ValueError(
            'The type of parameter is unknown. It should be either Parameter '
            'or Variable or ndarray, but the type was {}.'.format(
                type(parameter)))
    array = chainer.cuda.to_cpu(array)
    tensor = numpy_helper.from_array(array, context.get_name(parameter))
    return tensor


def rename_variable_name(
        context, variables, named_vars, new_names, prefix='Input'):
    # Update ``named_vars`` keys to ``new_names``
    if isinstance(variables, (list, tuple)):
        if new_names is None:
            new_names = ['{}_{}'.format(prefix, i)
                         for i in range(len(named_vars))]
        if not isinstance(new_names, (list, tuple)) or\
                len(variables) != len(new_names):
            raise ValueError(
                'Replacing name list is not match with input (or output) '
                'variables')
        for i, var in enumerate(variables):
            del named_vars[context.get_name(var)]
            new_name = new_names[i]
            named_vars[new_name] = var
            context.set_name(var, new_name, pinned=True)
    elif isinstance(variables, dict):
        if new_names is None:
            new_names = {k: '{}_{}'.format(prefix, i)
                         for i, k in enumerate(variables.keys())}
        if not isinstance(new_names, (list, tuple, dict)) or\
                len(variables) != len(new_names):
            raise ValueError(
                'Replacing name dict is not match with input (or output) '
                'variables')
        if isinstance(new_names, (list, tuple)):
            new_names = {k: v for k, v in zip(variables.keys(), new_names)}
        for k, v in variables.items():
            if k not in new_names:
                raise ValueError(
                    'Key of replacing name is not found in variables')
            del named_vars[context.get_name(v)]
            new_name = new_names[k]
            named_vars[new_name] = v
            context.set_name(v, new_name, pinned=True)
    elif isinstance(variables, chainer.Variable):
        if not new_names:
            new_names = prefix + '_0'
        if isinstance(new_names, (list, tuple)):
            if len(new_names) != 1:
                raise ValueError('Replacing name must be single')
            new_name = new_names[0]
        elif isinstance(new_names, str):
            new_name = new_names
        else:
            raise ValueError(
                'Type {} is not supported for single variable'.format(
                    type(new_name)))
        del named_vars[context.get_name(variables)]
        named_vars[new_name] = variables
        context.set_name(variables, new_name, pinned=True)


def format_customized_shapes(args, shapes):
    if isinstance(args, (list, tuple)):
        if not isinstance(shapes, list) or len(args) != len(shapes):
            raise ValueError('Customized shapes cannot fit for input list')
        for i, (arg, shape) in enumerate(zip(args, shapes)):
            if len(arg.shape) != len(shape):
                raise ValueError(
                    'Index-{} shape length must be same as input'.format(i))
        return shapes
    elif isinstance(args, dict):
        if not isinstance(shapes, (list, dict)) or\
                len(args) != len(shapes):
            raise ValueError('Customized shapes cannot fit for input dict')
        if isinstance(shapes, list):
            shapes = {k: v for k, v in zip(args.keys(), shapes)}
        formatted_shapes = []
        for k, arg in args.items():
            if k not in shapes:
                raise ValueError(
                    'Key "{}" is not found in customized shapes'.format(k))
            if len(arg.shape) != len(shapes[k]):
                raise ValueError(
                    'Key "{}" shape length must be same as input'.format(k))
            formatted_shapes.append(shapes[k])
        return formatted_shapes
    else:
        assert isinstance(args, (chainer.Variable, chainer.get_array_types()))
        if isinstance(shapes, list):
            if len(shapes) != 1:
                raise ValueError('Customized shape must be single')
        elif not isinstance(shapes, tuple):
            raise ValueError(
                'Type {} is not supported for single input'.format(
                    type(shapes)))
        else:
            shapes = [shapes]
        if len(args.shape) != len(shapes[0]):
            raise ValueError('Shape length must be same as input')
        return shapes


class RetainInputHook(chainer.LinkHook):
    """Retain temporary inputs

    Function nodes manage inputs variable nodes using weak reference. When
    variable is made as temporary value, exporter cannot get the corresponded
    variable from the variable node because the reference is collected. To
    resolve it, retain all inputs and will use when make computational graph.

    To reduce memory size, this hook retains only variables not showed in link
    inputs. To enable this feature, links are required to use ``forward``, not
    ``__call__``.
    """

    def __init__(self):
        self.link_inputs = set()
        self.retain_inputs = []
        self.replaced_inputs = []

        self.org_apply = chainer.function_node.FunctionNode.apply

        def hooked_apply(_self, inputs):
            ret = self.org_apply(_self, inputs)
            func_inodes = list(_self.inputs)
            for i, inode in enumerate(func_inodes):
                referenced_var = inode.get_variable_or_none()
                if referenced_var is None:
                    # This variable is created within function node and weakref
                    # is lost. Make temporary variable and retain it.
                    temp_var = chainer.as_variable(inputs[i])
                    func_inodes[i] = temp_var.node
                    self.retain_inputs.append(temp_var)
                else:
                    if id(referenced_var) not in self.link_inputs:
                        # This variable is created within link forward, outside
                        # of function node. To avoid to lose reference out
                        # of the forward, retain the variable.
                        self.retain_inputs.append(referenced_var)
            self.replaced_inputs.append((_self, _self.inputs))
            _self.inputs = tuple(func_inodes)
            return ret
        self.hooked_apply = hooked_apply

    def _extract_inputs(self, args):
        # Retain only chainer.Variable (and its collection)
        # Other type args are ignored and not checked instance IDs
        # If these variable are used in FunctionNode, they will be retained
        ret = set()
        if isinstance(args, chainer.Variable):
            ret.add(id(args))
        elif isinstance(args, (list, tuple)):
            for arg in args:
                ret |= self._extract_inputs(arg)
        elif isinstance(args, dict):
            for arg in args.values():
                ret |= self._extract_inputs(arg)
        return ret

    def forward_preprocess(self, args):
        self.link_inputs |= self._extract_inputs(args.args)
        self.link_inputs |= self._extract_inputs(args.kwargs)

    def forward_postprocess(self, args):
        self.link_inputs.clear()

    def __enter__(self):
        chainer.function_node.FunctionNode.apply = self.hooked_apply
        return super().__enter__()

    def __exit__(self, *exc_details):
        chainer.function_node.FunctionNode.apply = self.org_apply
        for _self, inputs in self.replaced_inputs:
            _self.inputs = inputs
        super().__exit__(*exc_details)


def export(model, args, filename=None, export_params=True,
           graph_name='Graph', save_text=False, opset_version=None,
           input_names=None, output_names=None, train=False,
           return_named_inout=False, external_converters=None,
           external_opset_imports=None, input_shapes=None):
    """Export function for chainer.Chain in ONNX format.

    This function performs a forward computation of the given
    :class:`~chainer.Chain`, ``model``, by passing the given arguments ``args``
    directly. It means, the output :class:`~chainer.Variable` object ``y`` to
    make the computational graph will be created by:

    ``y = model(*args)``

    ``external_converters`` and ``external_opset_imports`` are for external
    custom operator. When some ~chainer.FunctionNode are expected to convert to
    own customized operator, set converter function with ~chainer.FunctionNode
    name.

    >>> import onnx
    >>> def custom_converter(param):
    ...     return onnx.helper.make_node(
    ...         'CustomizedRelu', param.input_names, param.output_names,
    ...         domain='chainer'),
    >>>
    >>> external_converters = {'ReLU': custom_converter}
    >>> external_imports = {'chainer': 0}
    >>>
    >>> model = chainer.Sequential(F.relu)  # set the target model
    >>> args = chainer.Variable(np.random.rand(1,10))  # set dummy input
    >>> onnx_graph = onnx_chainer.export(
    ...     model, args,
    ...     external_converters=external_converters,
    ...     external_opset_imports=external_imports)

    Returned model has ``CustomizedRelu`` node.

    Args:
        model (~chainer.Chain): The model object you want to export in ONNX
            format. It should have :meth:`__call__` method because the second
            argument ``args`` is directly given to the model by the ``[]``
            accessor.
        args (list or dict): The arguments which are given to the model
            directly.
        filename (str or file-like object): The filename used for saving the
            resulting ONNX model. If None, nothing is saved to the disk.
        export_params (bool): If True, this function exports all the parameters
            included in the given model at the same time. If False, the
            exported ONNX model doesn't include any parameter values.
        graph_name (str): A string to be used for the ``name`` field of the
            graph in the exported ONNX model.
        save_text (bool): If True, the text format of the output ONNX model is
            also saved with ``.txt`` extention.
        opset_version (int): The operator set version of ONNX. If not specified
            or ``None`` is given, the latest opset version of the onnx module
            is used. If an integer is given, it will be ensured that all the
            operator version in the exported ONNX file is less than this value.
        input_names (str, list or dict): Customize input names of the graph.
            Number of ``input_names`` must be same as number of ``args``.
            When set dict type, keys must be same as ``args``'s keys.
        output_names (str, list or dict): Customize output name of the graph.
            Number of ``output_names`` must be same as actual outputs from
            ``model``. When set dict type, keys must be same as the key of
            ``model`` output.
        train (bool): If True, output computational graph with train mode.
        return_named_inout (bool): If set True, return ONNX model with named
            inputs, and named outputs.
        external_converters (dict): Add-on converter. Convert functions
            keyed by ~chainer.FunctionNode name.
        external_opset_imports (dict): Import external opset. opset version
            number keyed by domain name.
        input_shapes (tuple, list, dict): Input shape of output graph follows
            the customized shapes if set. When input are collection type, set
            list or dict. Tuple of tuple is not allowed.

    Returns:
        ~onnx.ModelProto or tuple:
            When ``return_named_inout`` is ``False``, return ModelProto as an
            ONNX model. Otherwise return the tuple of ModelProto, named inputs
            and outputs, both inputs and outputs are list of ~chainer.Variable.

    """

    _check_available()

    with chainer.using_config('train', train),\
            chainer.using_config('in_recomputing', True),\
            chainer.using_config('enable_backprop', True):
        return _export(
            model, args, filename, export_params, graph_name, save_text,
            opset_version, input_names, output_names, return_named_inout,
            external_converters, external_opset_imports, input_shapes)


def _export(model, args, filename, export_params, graph_name, save_text,
            opset_version, input_names, output_names, return_named_inout,
            external_converters, external_opset_imports, input_shapes):
    if opset_version is None:
        opset_version = min(
            int(onnx.defs.onnx_opset_version()), MAXIMUM_OPSET_VERSION)
    elif opset_version < MINIMUM_OPSET_VERSION or \
            opset_version > MAXIMUM_OPSET_VERSION:
        warnings.warn(
            'ONNX-Chainer has been tested only with opset_version {} ~ {}'
            'The ONNX file exported with your requested opset_version ({}) '
            'may cause some problems because the converters used for the '
            'opset_version have not been tested.'.format(
                MINIMUM_OPSET_VERSION, MAXIMUM_OPSET_VERSION, opset_version))

    if input_shapes is not None:
        # if input shapes are invalid, raise exception before forwarding.
        input_shapes = format_customized_shapes(args, input_shapes)

    with RetainInputHook():
        # Forward computation
        context = Context(model)
        network_inputs = OrderedDict()
        if isinstance(args, tuple):
            args = list(args)
        if isinstance(args, list):
            for i, arg in enumerate(args):
                if isinstance(arg, chainer.get_array_types()):
                    args[i] = chainer.Variable(arg)
                network_inputs[context.get_name(args[i])] = args[i]
            outputs = model(*args)
        elif isinstance(args, dict):
            for key, arg in args.items():
                if isinstance(arg, chainer.get_array_types()):
                    args[key] = chainer.Variable(arg)
                network_inputs[context.get_name(args[key])] = args[key]
            outputs = model(**args)
        elif isinstance(args, chainer.get_array_types()):
            args = chainer.Variable(args)
            network_inputs[context.get_name(args)] = args
            outputs = model(args)
        elif isinstance(args, chainer.Variable):
            network_inputs[context.get_name(args)] = args
            outputs = model(args)
        else:
            raise ValueError(
                'The \'args\' argument should be a list, tuple, dict, '
                'numpy array, or Chainer Variable. But a {} object was '
                'given.'.format(type(args)))
        rename_variable_name(context, args, network_inputs, input_names)

        initializers = []
        input_tensors = []
        param_names = set()
        for org_name, param in model.namedparams():
            # `model.namedparams()` has `include_uninit` flag but not use, to
            # output user warning
            if param.array is None:
                warnings.warn(
                    'The parameter \'{}\' is not initialized, skip setting to '
                    'ONNX graph'.format(org_name))
                continue
            name = context.get_name(param)
            param_names.add(name)
            tensor = convert_parameter(param, context)
            initializers.append(tensor)
            input_tensors.append(helper.make_tensor_value_info(
                name, tensor.data_type, tensor.dims))

        for i, (name, var) in enumerate(network_inputs.items()):
            shape = var.shape if input_shapes is None else input_shapes[i]
            input_tensors.append(helper.make_tensor_value_info(
                name, NP_TYPE_TO_TENSOR_TYPE[var.dtype], shape))

        if external_converters:
            chainer.utils.experimental('external_converters')
            converters = dict(mapping.converters, **external_converters)
        else:
            converters = mapping.converters

        if isinstance(outputs, (list, tuple)):
            flat_outputs = outputs
        elif isinstance(outputs, dict):
            flat_outputs = list(outputs.values())
        elif isinstance(outputs, chainer.Variable):
            flat_outputs = [outputs]
        else:
            raise RuntimeError(
                'Unexpected output type from the model: {}'.format(
                    type(outputs)))
        if not all([isinstance(o, chainer.Variable) for o in flat_outputs]):
            raise ValueError('The all \'outputs\' must be Chainer Variable')
        network_outputs = OrderedDict(
            [(context.get_name(var), var) for var in flat_outputs])
        if output_names:
            rename_variable_name(
                context, outputs, network_outputs, output_names)

        o = Graph(context, converters, opset_version,
                  param_names | set(network_inputs.keys()),
                  network_outputs)
        o.to_onnx_graph()

    implicit_input_names = set(context.implicit_inputs.keys())
    for name in implicit_input_names:
        tensor = convert_parameter(context.implicit_inputs[name], context)
        initializers.append(tensor)
        input_tensors.append(helper.make_tensor_value_info(
            name, tensor.data_type, tensor.dims))

    # If additional parameters are created during conversion
    for param in context.parameters:
        tensor = convert_parameter(param, context)
        initializers.append(tensor)
        input_tensors.append(helper.make_tensor_value_info(
            context.get_name(param), tensor.data_type, tensor.dims))

    # Convert output tensors
    output_tensors = []
    for name, var in network_outputs.items():
        output_tensors.append(helper.make_tensor_value_info(
            name, NP_TYPE_TO_TENSOR_TYPE[var.dtype], var.shape))

    if not export_params:
        initializers = []

    onnx_graph = helper.make_graph(
        o.graph, graph_name, input_tensors, output_tensors,
        initializer=initializers)

    opset_imports = [helper.make_operatorsetid('', opset_version)]
    if external_opset_imports:
        chainer.utils.experimental('external_opset_imports')
        for domain, version in external_opset_imports.items():
            opset_imports.append(helper.make_operatorsetid(domain, version))
    model = helper.make_model(
        onnx_graph,
        producer_name='Chainer',
        producer_version=chainer.__version__,
        opset_imports=opset_imports
    )

    model.ir_version = onnx.IR_VERSION
    check_onnx_model(model, external_converters, external_opset_imports)

    if input_shapes is not None:
        for output in model.graph.output:
            for d in output.type.tensor_type.shape.dim:
                d.Clear()
        model = shape_inference.infer_shapes(model)
        check_onnx_model(model, external_converters, external_opset_imports)

    if filename is not None and isinstance(filename, str):
        with open(filename, 'wb') as fp:
            fp.write(model.SerializeToString())
        if save_text:
            with open(filename + '.txt', 'w') as fp:
                print(model, file=fp)
    elif hasattr(filename, 'write'):
        filename.write(model.SerializeToString())

    if return_named_inout:
        chainer.utils.experimental('return_named_inout')
        return model, network_inputs, network_outputs
    return model


def check_onnx_model(onnx_model, external_converters, external_opset_imports):
    try:
        checker.check_model(onnx_model)
    except onnx.checker.ValidationError as e:
        if external_converters is None:
            raise e
        else:
            # ONNX version >= 1.5: default checker skips schema check when
            # non standard domain is set. In ONNX-Chainer, external ops without
            # doamin is also accepted, but show warning.
            # ONNX version < 1.5: the checker does not skip schema check
            # regardless domain is set or not. In ONNX-Chainer, ignore
            # errors when external ops are set.
            if is_support_non_standard_domain():
                if external_opset_imports:
                    raise e
                else:
                    warnings.warn(
                        'ValidationError is occurred but ignored. '
                        'ONNX-Chainer recommends to set '
                        '`external_opset_imports` when using '
                        '`external_converters` on exporting. Please take care '
                        'about ONNX format check is insufficient. Error '
                        'message:\n{}'.format(str(e)), UserWarning)
            else:
                warnings.warn(
                    'ValidationError is occurred but ignored because '
                    'exporting with `external_converters`. Please take care '
                    'about ONNX format check is insufficient. Error '
                    'message:\n{}'.format(str(e)), UserWarning)



import functools
from operator import mul

import numpy
import six

import chainer
from chainer import backend
from chainer.backends import cuda
from chainer.backends import intel64
from chainer import configuration
from chainer import function_node
from chainer.functions.pooling import max_pooling_nd_kernel
from chainer.functions.pooling import pooling_nd
from chainer.utils import conv_nd
import chainerx

if cuda.cudnn_enabled:
    _cudnn_version = cuda.cuda.cudnn.getVersion()


class MaxPoolingND(pooling_nd._PoolingND):

    """Max pooling over a set of N-dimensional planes.

    .. warning::

        This feature is experimental. The interface can change in the future.

    """

    def __init__(self, ndim, ksize, stride=None, pad=0, cover_all=True,
                 return_indices=False):
        super(MaxPoolingND, self).__init__(
            ndim, ksize, stride=stride, pad=pad, cover_all=cover_all,
            return_indices=return_indices)

    def forward_chainerx(self, x):
        ndim = self.ndim
        ksize = self.ksize
        stride = self.stride
        pad = self.pad
        cover_all = self.cover_all

        # TODO(sonots): Support return_indices in ChainerX
        if self.return_indices:
            return chainer.Fallback
        if x[0].device.backend.name == 'cuda':
            # TODO(sonots): Support more ndim in ChainerX
            if ndim not in [2, 3]:
                return chainer.Fallback
        y = chainerx.max_pool(x[0], ksize, stride, pad, cover_all)
        return y,

    def forward_cpu(self, x):
        if (self.ndim == 2
                and intel64.should_use_ideep('>=auto')
                and intel64.inputs_all_ready(x)):
            return self._forward_2d_ideep(x)

        ksize = self.ksize
        stride = self.stride
        pad = self.pad
        cover_all = self.cover_all

        in_shape = x[0].shape
        in_dtype = x[0].dtype

        col = conv_nd.im2col_nd_cpu(
            x[0], ksize, stride, pad,
            pval=-float('inf'),
            cover_all=cover_all)
        n, c = col.shape[:2]
        mid = (len(col.shape) - 2) // 2 + 2
        ksize = col.shape[2:mid]
        outs = col.shape[mid:]
        # (n, c, k_1 * k_2 * ... * k_N, out_1, out_2, ..., out_N)
        col_shape = (n, c) + (functools.reduce(mul, ksize),) + outs
        col = col.reshape(col_shape)

        # We select maximum twice, since the implementation using numpy.choose
        # hits its bug when kh * kw >= 32.
        y = col.max(axis=2)

        self._in_shape = in_shape
        self._in_dtype = in_dtype
        self.indexes = col.argmax(axis=2)
        return y,

    def _forward_2d_ideep(self, x):
        assert self.ndim == 2
        kh, kw = self.ksize
        sy, sx = self.stride
        ph, pw = self.pad
        cover_all = self.cover_all

        self._in_shape = x[0].shape
        self._in_dtype = x[0].dtype
        self.retain_inputs((0,))

        n, c, h, w = x[0].shape
        y_h = conv_nd.get_conv_outsize(h, kh, sy, ph, cover_all)
        assert y_h > 0, 'Height in the output should be positive.'
        y_w = conv_nd.get_conv_outsize(w, kw, sx, pw, cover_all)
        assert y_w > 0, 'Width in the output should be positive.'
        pd = sy * (y_h - 1) + kh - h - ph
        pr = sx * (y_w - 1) + kw - w - pw

        pp = intel64.ideep.pooling2DParam(
            (n, c, y_h, y_w),
            kh, kw,
            sy, sx,
            ph, pw,
            pd, pr,
            intel64.ideep.pooling2DParam.pooling_max)
        y, indexes = intel64.ideep.pooling2D.Forward(
            intel64.ideep.array(x[0]), pp)

        self.indexes = indexes
        return y,

    def forward_gpu(self, x):
        if chainer.should_use_cudnn('>=auto') and 2 <= self.ndim <= 3:
            # With cuDNN v3 or greater, use cuDNN implementation for inputs
            # with spatial dimensions of two or more.
            return self.forward_cudnn(x)

        ndim = self.ndim
        ksize = self.ksize
        stride = self.stride
        pad = self.pad
        cover_all = self.cover_all

        in_shape = x[0].shape
        in_dtype = x[0].dtype

        n, c = in_shape[:2]
        dims = in_shape[2:]
        ys = tuple(conv_nd.get_conv_outsize(d, k, s, p, cover_all)
                   for (d, k, s, p) in six.moves.zip(dims, ksize, stride, pad))
        # (n, c, y_1, y_2, ..., y_N)
        y_shape = (n, c) + ys
        y = cuda.cupy.empty(y_shape, dtype=x[0].dtype)
        indexes = cuda.cupy.empty(y_shape, dtype=numpy.int32)

        in_params, out_params, operation, name = \
            max_pooling_nd_kernel.MaxPoolingNDKernelForward.generate(ndim)
        cuda.elementwise(in_params, out_params, operation, name)(
            x[0].reduced_view(),
            *(dims + ys + ksize + stride + pad + (y, indexes)))

        self._in_shape = in_shape
        self._in_dtype = in_dtype
        self.indexes = indexes
        return y,

    def backward(self, indexes, gy):
        return MaxPoolingNDGrad(self).apply(gy)

    def get_cudnn_pool_mode(self):
        if _cudnn_version >= 6000 and configuration.config.cudnn_deterministic:
            return cuda.cuda.cudnn.CUDNN_POOLING_MAX_DETERMINISTIC
        else:
            return cuda.cuda.cudnn.CUDNN_POOLING_MAX


class MaxPoolingNDGrad(function_node.FunctionNode):

    def __init__(self, func):
        self.func = func

    def forward_cpu(self, gy):
        func = self.func

        if (func.ndim == 2
                and intel64.should_use_ideep('>=auto')
                and intel64.inputs_all_ready(gy)):
            return self._forward_2d_ideep(gy)

        ndim = func.ndim
        ksize = func.ksize
        stride = func.stride
        pad = func.pad
        in_shape = func._in_shape
        in_dtype = func._in_dtype
        indexes = func.indexes

        n, c = gy[0].shape[:2]
        outs = gy[0].shape[2:]
        dims = in_shape[2:]
        prod_outs = functools.reduce(mul, outs)
        prod_ksize = functools.reduce(mul, ksize)

        gcol = numpy.zeros(n * c * prod_outs * prod_ksize, dtype=in_dtype)

        indexes = (
            indexes.flatten()
            + numpy.arange(0, indexes.size * prod_ksize, prod_ksize))

        gcol[indexes] = gy[0].ravel()
        gcol_shape = (n, c) + outs + ksize
        gcol = gcol.reshape(gcol_shape)
        for i in six.moves.range(ndim):
            gcol = numpy.swapaxes(gcol, 2 + i, ndim + 2 + i)

        gx = conv_nd.col2im_nd_cpu(gcol, stride, pad, dims)
        return gx,

    def _forward_2d_ideep(self, gy):
        func = self.func

        # FIXME
        # Here we expect indexes is returned from MKL-DNN
        # otherwise, there are dtype mismatch for reorder (int64-->uint8)
        if not isinstance(func.indexes, intel64.ideep.mdarray):
            return self.forward_cpu(gy)

        kh, kw = func.ksize
        sy, sx = func.stride
        ph, pw = func.pad
        indexes = func.indexes
        in_shape = func._in_shape

        n, c, h, w = in_shape
        y_h, y_w = gy[0].shape[2:]
        x = func.get_retained_inputs()[0].array

        pd = sy * (y_h - 1) + kh - h - ph
        pr = sx * (y_w - 1) + kw - w - pw

        pp = intel64.ideep.pooling2DParam(
            func._in_shape,
            kh, kw,
            sy, sx,
            ph, pw,
            pd, pr,
            intel64.ideep.pooling2DParam.pooling_max)

        indexes = intel64.ideep.array(indexes)
        gx = intel64.ideep.pooling2D.Backward(
            intel64.ideep.array(x),
            intel64.ideep.array(gy[0]),
            indexes, pp)
        return gx,

    def forward_gpu(self, gy):
        func = self.func

        if func.is_cudnn_used:
            return func.backward_cudnn(gy)

        ndim = func.ndim
        ksize = func.ksize
        stride = func.stride
        pad = func.pad
        in_shape = func._in_shape
        in_dtype = func._in_dtype
        indexes = backend.from_chx(func.indexes)

        n, c = in_shape[:2]
        dims = in_shape[2:]
        ys = gy[0].shape[2:]
        gx = cuda.cupy.empty(in_shape, in_dtype)

        in_params, out_params, operation, name = \
            max_pooling_nd_kernel.MaxPoolingNDKernelBackward.generate(ndim)
        cuda.elementwise(in_params, out_params, operation, name)(
            gy[0].reduced_view(), indexes.reduced_view(),
            *(dims + ys + ksize + stride + pad + (gx,)))
        return gx,

    def backward(self, indexes, ggx):
        return MaxPoolingNDWithIndexes(self.func).apply(ggx)


class MaxPoolingNDWithIndexes(function_node.FunctionNode):

    def __init__(self, func):
        self.func = func

    def forward_cpu(self, x):
        func = self.func
        ndim = func.ndim
        ksize = func.ksize
        stride = func.stride
        pad = func.pad
        cover_all = func.cover_all
        indexes = backend.from_chx(func.indexes)

        col = conv_nd.im2col_nd_cpu(
            x[0], ksize, stride, pad,
            pval=-float('inf'),
            cover_all=cover_all)
        n, c = col.shape[:2]
        mid = (len(col.shape) - 2) // 2 + 2
        ksize = col.shape[2:mid]
        outs = col.shape[mid:]
        # (n, c, k_1 * k_2 * ... * k_N, out_1, out_2, ..., out_N)
        ksize_total = functools.reduce(mul, ksize)
        col_shape = (n, c) + (ksize_total,) + outs
        col = col.reshape(col_shape)
        # (n, c, out_1, ..., out_N, k_1 * .. * k_N)
        col_indexes = (0, 1) + tuple(six.moves.range(3, 3 + ndim)) + (2,)
        col = col.transpose(col_indexes)
        col = col.reshape(-1, ksize_total)

        indexes = indexes.ravel()
        col = col[numpy.arange(len(indexes)), indexes]
        return col.reshape((n, c) + outs),

    def forward_gpu(self, inputs):
        func = self.func

        if func.is_cudnn_used:
            x = func.get_retained_inputs()[0].array
            return self._forward_gpu_compute_indexes_again((x, inputs[0]))

        ndim = func.ndim
        ksize = func.ksize
        stride = func.stride
        pad = func.pad
        cover_all = func.cover_all
        indexes = backend.from_chx(func.indexes)

        x, = inputs
        in_shape = x.shape
        in_dtype = x.dtype

        n, c = in_shape[:2]
        dims = in_shape[2:]

        ys = tuple(conv_nd.get_conv_outsize(d, k, s, p, cover_all)
                   for (d, k, s, p) in six.moves.zip(dims, ksize, stride, pad))
        # (n, c, y_1, y_2, ..., y_N)
        y_shape = (n, c) + ys
        y = cuda.cupy.empty(y_shape, dtype=x.dtype)

        cls = max_pooling_nd_kernel.MaxPoolingNDKernelForwardWithIndexes
        in_params, out_params, operation, name = cls.generate(ndim)
        cuda.elementwise(in_params, out_params, operation, name)(
            x.reduced_view(),
            *(dims + ys + ksize + stride + pad + (indexes.reduced_view(), y)))

        self._in_shape = in_shape
        self._in_dtype = in_dtype
        return y,

    def _forward_gpu_compute_indexes_again(self, inputs):
        func = self.func
        ndim = func.ndim
        ksize = func.ksize
        stride = func.stride
        pad = func.pad
        cover_all = func.cover_all

        x, ggx = inputs
        in_shape = x.shape
        in_dtype = x.dtype

        n, c = in_shape[:2]
        dims = in_shape[2:]

        ys = tuple(conv_nd.get_conv_outsize(d, k, s, p, cover_all)
                   for (d, k, s, p) in six.moves.zip(dims, ksize, stride, pad))
        # (n, c, y_1, y_2, ..., y_N)
        y_shape = (n, c) + ys
        y = cuda.cupy.empty(y_shape, dtype=x.dtype)

        cls = max_pooling_nd_kernel.MaxPoolingNDKernelForwardWithIndexes1
        in_params, out_params, operation, name = cls.generate(ndim)
        cuda.elementwise(in_params, out_params, operation, name)(
            x.reduced_view(),
            *(dims + ys + ksize + stride + pad + (ggx.reduced_view(), y)))

        self._in_shape = in_shape
        self._in_dtype = in_dtype
        return y,


def max_pooling_nd(x, ksize, stride=None, pad=0, cover_all=True,
                   return_indices=False):
    """N-dimensionally spatial max pooling function.

    .. warning::

        This feature is experimental. The interface can change in the future.

    This function provides a N-dimensionally generalized version of
    :func:`~chainer.functions.max_pooling_2d`. This acts similarly to
    :func:`~chainer.functions.convolution_nd`, but it computes the maximum of
    input spatial patch for each channel without any parameter instead of
    computing the inner products.

    Args:
        x (~chainer.Variable): Input variable.
        ksize (int or tuple of ints): Size of pooling window. ``ksize=k`` and
            ``ksize=(k, k, ..., k)`` are equivalent.
        stride (int or tuple of ints or None): Stride of pooling applications.
            ``stride=s`` and ``stride=(s,s, ..., s)`` are equivalent. If
            ``None`` is specified, then it uses same stride as the pooling
            window size.
        pad (int or tuple of ints): Spatial padding width for the input array.
            ``pad=p`` and ``pad=(p, p, ..., p)`` are equivalent.
        cover_all (bool): If ``True``, all spatial locations are pooled into
            some output pixels. It may make the output size larger.
        return_indices (bool): If ``True``, pooling indices array is returned
            together with the output variable. The returned indices are
            expected for use by :func:`chainer.functions.upsampling_nd`.
            Note that cuDNN will not be used for this function if
            ``return_indices`` is set to ``True``, as cuDNN does not return
            indices information.

    Returns:
        ~chainer.Variable or tuple:
            When ``return_indices`` is ``False`` (default), returns the output
            variable.
            When ``True``, returns the tuple of the output variable and
            pooling indices (:ref:`ndarray`). Pooling indices will be on the
            same device as the input.

    """
    ndim = len(x.shape[2:])

    func = MaxPoolingND(ndim, ksize, stride, pad, cover_all, return_indices)
    if return_indices:
        with chainer.using_config('use_cudnn', 'never'):
            out = func.apply((x,))[0]
        return out, func.indexes

    return func.apply((x,))[0]


def max_pooling_1d(x, ksize, stride=None, pad=0, cover_all=True,
                   return_indices=False):
    """1-dimensional spatial max pooling function.

    .. warning::

        This feature is experimental. The interface can change in the future.

    .. note::

        This function calls :func:`~chainer.functions.max_pooling_nd`
        internally, so see the details of the behavior in
        the documentation of :func:`~chainer.functions.max_pooling_nd`.

    """
    if len(x.shape[2:]) != 1:
        raise ValueError(
            'The number of dimensions under channel dimension of the input '
            '\'x\' should be 1. But the actual ndim was {}.'.format(
                len(x.shape[2:])))
    return max_pooling_nd(x, ksize, stride, pad, cover_all, return_indices)


def max_pooling_2d(x, ksize, stride=None, pad=0, cover_all=True,
                   return_indices=False):
    """Spatial max pooling function.

    This function acts similarly to :func:`~chainer.functions.convolution_2d`,
    but it computes the maximum of input spatial patch for each channel without
    any parameter instead of computing the inner products.

    Args:
        x (~chainer.Variable): Input variable.
        ksize (int or pair of ints): Size of pooling window. ``ksize=k`` and
            ``ksize=(k, k)`` are equivalent.
        stride (int or pair of ints or None): Stride of pooling applications.
            ``stride=s`` and ``stride=(s, s)`` are equivalent. If ``None`` is
            specified, then it uses same stride as the pooling window size.
        pad (int or pair of ints): Spatial padding width for the input array.
            ``pad=p`` and ``pad=(p, p)`` are equivalent.
        cover_all (bool): If ``True``, all spatial locations are pooled into
            some output pixels. It may make the output size larger.
        return_indices (bool): If ``True``, pooling indices array is returned
            together with the output variable. The returned indices are
            expected for use by :func:`chainer.functions.upsampling_2d`.
            Note that cuDNN will not be used for this function if
            ``return_indices`` is set to ``True``, as cuDNN does not return
            indices information.

    Returns:
        ~chainer.Variable or tuple:
            When ``return_indices`` is ``False`` (default), returns the output
            variable.
            When ``True``, returns the tuple of the output variable and
            pooling indices (:ref:`ndarray`). Pooling indices will be on the
            same device as the input.

    """
    if len(x.shape[2:]) != 2:
        raise ValueError(
            'The number of dimensions under channel dimension of the input '
            '\'x\' should be 2. But the actual ndim was {}.'.format(
                len(x.shape[2:])))
    return max_pooling_nd(x, ksize, stride, pad, cover_all, return_indices)


def max_pooling_3d(x, ksize, stride=None, pad=0, cover_all=True,
                   return_indices=False):
    """3-dimensional spatial max pooling function.

    .. warning::

        This feature is experimental. The interface can change in the future.

    .. note::

        This function calls :func:`~chainer.functions.max_pooling_nd`
        internally, so see the details of the behavior in
        the documentation of :func:`~chainer.functions.max_pooling_nd`.

    """
    if len(x.shape[2:]) != 3:
        raise ValueError(
            'The number of dimensions under channel dimension of the input '
            '\'x\' should be 3. But the actual ndim was {}.'.format(
                len(x.shape[2:])))
    return max_pooling_nd(x, ksize, stride, pad, cover_all, return_indices)



import functools
import warnings

import numpy

import chainer
from chainer import configuration
from chainer import functions
from chainer import link
import chainer.serializer as serializer_mod
from chainer.utils import argument


class DecorrelatedBatchNormalization(link.Link):

    """Decorrelated batch normalization layer.

    This link wraps the
    :func:`~chainer.functions.decorrelated_batch_normalization` and
    :func:`~chainer.functions.fixed_decorrelated_batch_normalization`
    functions. It works on outputs of linear or convolution functions.

    It runs in three modes: training mode, fine-tuning mode, and testing mode.

    In training mode, it normalizes the input by *batch statistics*. It also
    maintains approximated population statistics by moving averages, which can
    be used for instant evaluation in testing mode.

    In fine-tuning mode, it accumulates the input to compute *population
    statistics*. In order to correctly compute the population statistics, a
    user must use this mode to feed mini-batches running through whole training
    dataset.

    In testing mode, it uses pre-computed population statistics to normalize
    the input variable. The population statistics is approximated if it is
    computed by training mode, or accurate if it is correctly computed by
    fine-tuning mode.

    Args:
        size (int or tuple of ints): Size (or shape) of channel
            dimensions.
        groups (int): Number of groups to use for group whitening.
        decay (float): Decay rate of moving average
            which is used during training.
        eps (float): Epsilon value for numerical stability.
        dtype (numpy.dtype): Type to use in computing.

    See: `Decorrelated Batch Normalization <https://arxiv.org/abs/1804.08450>`_

    .. seealso::
       :func:`~chainer.functions.decorrelated_batch_normalization`,
       :func:`~chainer.functions.fixed_decorrelated_batch_normalization`

    Attributes:
        avg_mean (:ref:`ndarray`): Population mean.
        avg_projection (:ref:`ndarray`): Population
            projection.
        groups (int): Number of groups to use for group whitening.
        N (int): Count of batches given for fine-tuning.
        decay (float): Decay rate of moving average
            which is used during training.
        ~DecorrelatedBatchNormalization.eps (float): Epsilon value for
            numerical stability. This value is added to the batch variances.

    """

    def __init__(self, size, groups=16, decay=0.9, eps=2e-5,
                 dtype=numpy.float32):
        super(DecorrelatedBatchNormalization, self).__init__()
        C = size // groups
        self.avg_mean = numpy.zeros((groups, C), dtype=dtype)
        self.register_persistent('avg_mean')
        avg_projection = numpy.zeros((groups, C, C), dtype=dtype)
        arange_C = numpy.arange(C)
        avg_projection[:, arange_C, arange_C] = 1
        self.avg_projection = avg_projection
        self.register_persistent('avg_projection')
        self.N = 0
        self.register_persistent('N')
        self.decay = decay
        self.eps = eps
        self.groups = groups

    def serialize(self, serializer):
        if isinstance(serializer, serializer_mod.Deserializer):
            serializer = _PatchedDeserializer(serializer, {
                'avg_mean': functools.partial(
                    fix_avg_mean, groups=self.groups),
                'avg_projection': functools.partial(
                    fix_avg_projection, groups=self.groups),
            })
        super(DecorrelatedBatchNormalization, self).serialize(serializer)

    def forward(self, x, **kwargs):
        """forward(self, x, *, finetune=False)

        Invokes the forward propagation of DecorrelatedBatchNormalization.

        In training mode, the DecorrelatedBatchNormalization computes moving
        averages of the mean and projection for evaluation during training,
        and normalizes the input using batch statistics.

        Args:
            x (:class:`~chainer.Variable`): Input variable.
            finetune (bool): If it is in the training mode and ``finetune`` is
                ``True``, DecorrelatedBatchNormalization runs in fine-tuning
                mode; it accumulates the input array to compute population
                statistics for normalization, and normalizes the input using
                batch statistics.

        """
        finetune, = argument.parse_kwargs(kwargs, ('finetune', False))

        if configuration.config.train:
            if finetune:
                self.N += 1
                decay = 1. - 1. / self.N
            else:
                decay = self.decay

            avg_mean = self.avg_mean
            avg_projection = self.avg_projection

            if configuration.config.in_recomputing:
                # Do not update statistics when extra forward computation is
                # called.
                if finetune:
                    self.N -= 1
                avg_mean = None
                avg_projection = None

            ret = functions.decorrelated_batch_normalization(
                x, groups=self.groups, eps=self.eps,
                running_mean=avg_mean, running_projection=avg_projection,
                decay=decay)
        else:
            # Use running average statistics or fine-tuned statistics.
            mean = self.avg_mean
            projection = self.avg_projection
            ret = functions.fixed_decorrelated_batch_normalization(
                x, mean, projection, groups=self.groups)
        return ret

    def start_finetuning(self):
        """Resets the population count for collecting population statistics.

        This method can be skipped if it is the first time to use the
        fine-tuning mode. Otherwise, this method should be called before
        starting the fine-tuning mode again.

        """
        self.N = 0


class _PatchedDeserializer(serializer_mod.Deserializer):

    def __init__(self, base, patches):
        self.base = base
        self.patches = patches

    def __repr__(self):
        return '_PatchedDeserializer({}, {})'.format(
            repr(self.base), repr(self.patches))

    def __call__(self, key, value):
        if key not in self.patches:
            return self.base(key, value)
        arr = self.base(key, None)
        arr = self.patches[key](arr)
        if value is None:
            return arr
        chainer.backend.copyto(value, arr)
        return value


def _warn_old_model():
    msg = (
        'Found moving statistics of old DecorrelatedBatchNormalization, whose '
        'algorithm was different from the paper.')
    warnings.warn(msg)


def fix_avg_mean(avg_mean, groups):
    if avg_mean.ndim == 2:  # OK
        return avg_mean
    elif avg_mean.ndim == 1:  # Issue #7706
        if groups != 1:
            _warn_old_model()
        return _broadcast_to(avg_mean, (groups,) + avg_mean.shape)
    raise ValueError('unexpected shape of avg_mean')


def fix_avg_projection(avg_projection, groups):
    if avg_projection.ndim == 3:  # OK
        return avg_projection
    elif avg_projection.ndim == 2:  # Issue #7706
        if groups != 1:
            _warn_old_model()
        return _broadcast_to(
            avg_projection, (groups,) + avg_projection.shape)
    raise ValueError('unexpected shape of avg_projection')


def _broadcast_to(array, shape):
    if hasattr(numpy, 'broadcast_to'):
        return numpy.broadcast_to(array, shape)
    else:
        # numpy 1.9 doesn't support broadcast_to method
        dummy = numpy.empty(shape)
        bx, _ = numpy.broadcast_arrays(array, dummy)
        return bx



import chainer

from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


@support((1, 6, 7))
def convert_Dropout(func, opset_version, input_names, output_names, context):
    if opset_version == 1:
        return onnx_helper.make_node(
            'Dropout', input_names, output_names,
            is_test=0 if chainer.config.train else 1,
            ratio=func.dropout_ratio,
            consumed_inputs=[1]
        ),
    elif opset_version == 6:
        return onnx_helper.make_node(
            'Dropout', input_names, output_names,
            is_test=0 if chainer.config.train else 1,
            ratio=func.dropout_ratio,
        ),
    elif opset_version == 7:
        return onnx_helper.make_node(
            'Dropout', input_names, output_names,
            ratio=func.dropout_ratio,
        ),



from onnx_chainer.export import convert_parameter  # NOQA
from onnx_chainer.export import export  # NOQA

from onnx_chainer.export import MAXIMUM_OPSET_VERSION  # NOQA
from onnx_chainer.export import MINIMUM_OPSET_VERSION  # NOQA

from onnx_chainer.export_testcase import export_testcase  # NOQA



import chainer
from chainer.dataset.tabular import tabular_dataset


def from_data(data, *, size=None):
    """Create a TabularDataset from lists/arrays/callables.

    >>> from chainer.dataset import tabular
    >>>
    >>> dataset = tabular.from_data([0, 1, 2])
    >>> dataset[0]
    0
    >>> dataset = tabular.from_data(([0, 1, 2], [3, 4, 5]))
    >>> dataset[0]
    (0, 3)
    >>> dataset = tabular.from_data((('a', [0, 1, 2]), ('b', [3, 4, 5])))
    >>> dataset.keys
    ('a', 'b')
    >>> dataset[0]
    (0, 3)
    >>> dataset = tabular.from_data({'a': [0, 1, 2], 'b': [3, 4, 5]})
    >>> sorted(dataset[0].items())
    [('a', 0), ('b', 3)]
    >>> dataset = tabular.from_data(('a', lambda i: i * i), size=10)
    >>> dataset[5]
    25

    Args:
        data (list, array, tuple, or dict): Data in following format.

            - `list/array`
            - `(str, list/array/callable)`
            - `((str, ...), callable)`
            - `((list/array)/(str, list/array/callable) \
                /((key, ...), callable), ...)`
            - `{str: (list/array/callable)/(str, ...): callable, ...}`
        size (int): The length of the dataset.
            This argument is required \
            when no lists/arrays exist in :obj:`data`.
    Return:
        A :class:`~chainer.dataset.TabularDataset`.
    """

    if isinstance(data, tuple):
        if len(data) == 2:
            key, d = data
            if isinstance(key, str):
                return _make_dataset(key, d, size)
            if isinstance(key, tuple) and all(isinstance(k, str) for k in key):
                return _make_dataset(key, d, size)

        for d in data:
            if isinstance(d, tuple):
                _, d = d

            if size is None:
                try:
                    size = len(d)
                except TypeError:
                    pass

        datasets = []
        for d in data:
            if isinstance(d, tuple):
                key, d = d
            else:
                key = None
            datasets.append(_make_dataset(key, d, size))

        return datasets[0].join(*datasets[1:]).astuple()

    elif isinstance(data, dict):
        for d in data.values():
            if size is None:
                try:
                    size = len(d)
                except TypeError:
                    pass

        datasets = []
        for key, d in data.items():
            datasets.append(_make_dataset(key, d, size))

        return datasets[0].join(*datasets[1:]).asdict()

    else:
        return _make_dataset(None, data, size)


def _make_dataset(key, data, size):
    if isinstance(data, chainer.get_array_types()):
        if key is None:
            key = '_{}'.format(id(data))
        return _Array(key, data)
    elif isinstance(data, list):
        if key is None:
            key = '_{}'.format(id(data))
        return _List(key, data)
    elif callable(data):
        if key is None:
            raise ValueError('key(s) must be specified for callable')
        if size is None:
            raise ValueError('size must be specified for callable')
        return _Index(size).transform(key, data)


class _Array(tabular_dataset.TabularDataset):

    def __init__(self, key, data):
        self._key = key
        self._data = data

    def __len__(self):
        return len(self._data)

    @property
    def keys(self):
        return self._key,

    @property
    def mode(self):
        return None

    def get_examples(self, indices, key_indices):
        if key_indices is None:
            key_indices = 0,

        if indices is None:
            return (self._data,) * len(key_indices)
        else:
            return (self._data[indices],) * len(key_indices)


class _List(tabular_dataset.TabularDataset):

    def __init__(self, key, data):
        self._key = key
        self._data = data

    def __len__(self):
        return len(self._data)

    @property
    def keys(self):
        return self._key,

    @property
    def mode(self):
        return None

    def get_examples(self, indices, key_indices):
        if key_indices is None:
            key_indices = 0,

        if indices is None:
            return (self._data,) * len(key_indices)
        elif isinstance(indices, slice):
            return (self._data[indices],) * len(key_indices)
        else:
            return ([self._data[index] for index in indices],) \
                * len(key_indices)


class _Index(tabular_dataset.TabularDataset):

    def __init__(self, size):
        self._len = size

    def __len__(self):
        return self._len

    @property
    def keys(self):
        return 'index',

    @property
    def mode(self):
        return None

    def get_examples(self, indices, key_indices):
        if indices is None:
            indices = slice(None)
        if isinstance(indices, slice):
            start, stop, step = indices.indices(len(self))
            indices = list(range(start, stop, step))

        if key_indices is None:
            key_indices = 0,

        return (indices,) * len(key_indices)



from chainer.backends import cuda
from chainer import function_node
from chainer import utils
from chainer.utils import type_check


_zeta_cpu = None


class Zeta(function_node.FunctionNode):

    def __init__(self, x):
        self._x = x

    @property
    def label(self):
        return 'zeta'

    def check_type_forward(self, in_types):
        type_check._argname(in_types, ('q'))
        q_type, = in_types
        type_check.expect(
            q_type.dtype.kind == 'f'
        )

    def forward_cpu(self, inputs):
        q, = inputs
        global _zeta_cpu
        if _zeta_cpu is None:
            try:
                from scipy import special
                _zeta_cpu = special.zeta
            except ImportError:
                raise ImportError('Scipy is not available. Forward computation'
                                  ' of zeta cannot be done.')
        self.retain_inputs((0,))
        return utils.force_array(_zeta_cpu(self._x, q), dtype=q.dtype),

    def forward_gpu(self, inputs):
        q, = inputs
        self.retain_inputs((0,))
        return utils.force_array(
            cuda.cupyx.scipy.special.zeta(self._x, q), dtype=q.dtype),

    def backward(self, indexes, gy):
        q, = self.get_retained_inputs()
        return gy[0] * -self._x * zeta(self._x + 1, q),


def zeta(x, q):
    """Zeta function.

    Differentiable only with respect to q

    .. note::
       Forward computation in CPU can not be done if
       `SciPy <https://www.scipy.org/>`_ is not available.

    Args:
        x (:class:`~chainer.Variable` or :ref:`ndarray`): Input variable.
        q (:class:`~chainer.Variable` or :ref:`ndarray`): Input variable.

    Returns:
        ~chainer.Variable: Output variable.
    """
    return Zeta(x).apply((q,))[0]



import numpy as np

from chainermn.communicators import _memory_utility
from chainermn.communicators import mpi_communicator_base


class FlatCommunicator(mpi_communicator_base.MpiCommunicatorBase):

    def __init__(self, mpi_comm):
        super(FlatCommunicator, self).__init__(mpi_comm)

        self.gpu_buffer_a = _memory_utility.DeviceMemory()
        self.gpu_buffer_b = _memory_utility.DeviceMemory()

    def multi_node_mean_grad(self, model, zero_fill=False):
        params = _memory_utility.extract_params_set_grad(model, zero_fill)
        itemsize = 4
        n_elems_total = _memory_utility.count_grad_elements(params,
                                                            zero_fill)
        n_bytes_total = n_elems_total * itemsize
        self.gpu_buffer_a.assign(n_bytes_total)
        self.gpu_buffer_b.assign(n_bytes_total)

        allreduce_grad_dtype = np.float32

        self._pack_params_to_buffer(params, 'grad', buffer=self.gpu_buffer_a,
                                    allreduce_grad_dtype=allreduce_grad_dtype,
                                    zero_fill=zero_fill)

        self._multi_node_mean(self.gpu_buffer_a.array(n_elems_total),
                              self.gpu_buffer_b.array(n_elems_total))

        self._unpack_params_from_buffer(params, 'grad', self.gpu_buffer_b,
                                        allreduce_grad_dtype, zero_fill)
